# -*- coding: utf-8 -*-

"""Template model."""


#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import logging
import os
import os.path as op
from operator import itemgetter
from pathlib import Path
import shutil

import numpy as np
# from numpy.lib.format import open_memmap
import scipy.io as sio
# from tqdm import tqdm

from .array import _index_of, _spikes_in_clusters, _spikes_per_cluster, SpikeSelector
from .traces import (
    get_ephys_reader, RandomEphysReader, extract_waveforms,
    get_spike_waveforms, export_waveforms)
from phylib.utils import Bunch
from phylib.utils._misc import _write_tsv_simple, read_tsv, read_python
from phylib.utils.geometry import linear_positions

logger = logging.getLogger(__name__)


#------------------------------------------------------------------------------
# Utility functions
#------------------------------------------------------------------------------

def read_array(path, mmap_mode=None):
    """Read a binary array in npy or mat format, avoiding nan and inf values."""
    path = Path(path)
    arr_name = path.name
    ext = path.suffix
    if ext == '.mat':  # pragma: no cover
        out = sio.loadmat(path)[arr_name]
    elif ext == '.npy':
        out = np.load(path, mmap_mode=mmap_mode)
    # Filter out nan and inf values.
    # NOTE: do not check for nan/inf values on mmap arrays.
    # TODO: virtual memmap array where the replacement is done on-the-fly when reading the array.
    if mmap_mode is None:
        for w in ('nan', 'inf'):
            errors = getattr(np, 'is' + w)(out)
            if np.any(errors):
                n = np.sum(errors)
                n_tot = errors.size
                logger.warning("%d/%d values are %s in %s, replacing by zero.", n, n_tot, w, path)
                out[errors] = 0
    return out


def write_array(name, arr):
    """Save an array to a binary file."""
    np.save(name, arr)


def from_sparse(data, cols, channel_ids):
    """Convert a sparse structure into a dense one.

    Parameters
    ----------

    data : array-like
        A (n_spikes, n_channels_loc, ...) array with the data.
    cols : array-like
        A (n_spikes, n_channels_loc) array with the channel indices of
        every row in data.
    channel_ids : array-like
        List of requested channel ids (columns).

    """
    # The axis in the data that contains the channels.
    if len(channel_ids) != len(np.unique(channel_ids)):
        raise NotImplementedError("Multiple identical requested channels "
                                  "in from_sparse().")
    channel_axis = 1
    shape = list(data.shape)
    assert data.ndim >= 2
    assert cols.ndim == 2
    assert data.shape[:2] == cols.shape
    n_spikes, n_channels_loc = shape[:2]
    # NOTE: we ensure here that `col` contains integers.
    c = cols.flatten().astype(np.int32)
    # Remove columns that do not belong to the specified channels.
    c[~np.isin(c, channel_ids)] = -1
    assert np.all(np.isin(c, np.r_[channel_ids.astype(np.int32), -1]))
    # Convert column indices to relative indices given the specified
    # channel_ids.
    cols_loc = _index_of(c, np.r_[channel_ids.astype(np.int32), -1]).reshape(cols.shape)
    assert cols_loc.shape == (n_spikes, n_channels_loc)
    n_channels = len(channel_ids)
    # Shape of the output array.
    out_shape = shape
    # The channel dimension contains the number of requested channels.
    # The last column contains irrelevant values.
    out_shape[channel_axis] = n_channels + 1
    out = np.zeros(out_shape, dtype=data.dtype)
    x = np.tile(np.arange(n_spikes)[:, np.newaxis],
                (1, n_channels_loc))
    assert x.shape == cols_loc.shape == data.shape[:2]
    out[x, cols_loc, ...] = data
    # Remove the last column with values outside the specified
    # channels.
    out = out[:, :-1, ...]
    return out


def load_metadata(filename):
    """Load cluster metadata from a TSV file.

    Return {field_name: dictionary cluster_id => value}.

    """
    data = read_tsv(filename)
    if not data:  # pragma: no cover
        return {}
    out = {}
    for d in data:
        if 'cluster_id' in d:
            cluster_id = d['cluster_id']
            for field, value in d.items():
                if field != 'cluster_id':
                    if field not in out:
                        out[field] = {}
                    out[field][cluster_id] = value
    return out


def save_metadata(filename, field_name, metadata):
    """Save metadata in a CSV file."""
    return _write_tsv_simple(filename, field_name, metadata)


#------------------------------------------------------------------------------
# Channel util functions
#------------------------------------------------------------------------------

def _all_positions_distinct(positions):
    """Return whether all positions are distinct."""
    return len(set(tuple(row) for row in positions)) == len(positions)


def get_closest_channels(channel_positions, channel_index, n=None):
    """Get the channels closest to a given channel on the probe."""
    x = channel_positions[:, 0]
    y = channel_positions[:, 1]
    x0, y0 = channel_positions[channel_index]
    d = (x - x0) ** 2 + (y - y0) ** 2
    out = np.argsort(d)
    if n:
        out = out[:n]
    assert out[0] == channel_index
    return out


#------------------------------------------------------------------------------
# PC computation
#------------------------------------------------------------------------------

def _compute_pcs(x, npcs):
    """Compute the PCs of an array x, where each row is an observation.
    x can be a 2D or 3D array. In the latter case, the PCs are computed
    and concatenated iteratively along the last axis."""

    # Ensure x is a 3D array.
    assert x.ndim == 3
    # Ensure double precision
    x = x.astype(np.float64)

    nspikes, nsamples, nchannels = x.shape

    cov_reg = np.eye(nsamples)
    assert cov_reg.shape == (nsamples, nsamples)

    pcs_list = []
    # Loop over channels
    for channel in range(nchannels):
        x_channel = x[:, :, channel]
        # Compute cov matrix for the channel
        assert x_channel.ndim == 2
        # Don't compute the cov matrix if there are no unmasked spikes
        # on that channel.
        alpha = 1. / nspikes
        if x_channel.shape[0] <= 1:
            cov = alpha * cov_reg
        else:
            cov_channel = np.cov(x_channel, rowvar=0)
            assert cov_channel.shape == (nsamples, nsamples)
            cov = alpha * cov_reg + cov_channel
        # Compute the eigenelements
        vals, vecs = np.linalg.eigh(cov)
        pcs = vecs.T.astype(np.float32)[np.argsort(vals)[::-1]]
        # Take the first npcs components.
        pcs_list.append(pcs[:npcs, ...])
    # Return the concatenation of the PCs on all channels, along the 3d axis,
    # except if there is only one element in the 3d axis. In this case
    # we convert to a 2D array.
    pcs = np.dstack(pcs_list)
    assert pcs.ndim == 3
    return pcs


def _project_pcs(x, pcs):
    """Project data points onto principal components.
    Arguments:
      * x: a 2D array.
      * pcs: the PCs as returned by `compute_pcs`.
    """
    assert x.ndim == 3
    assert pcs.ndim == 3
    features = []
    # for i, waveform in enumerate(x):
    #     assert waveform.ndim == 2
    #     # pcs.shape is (3, n_samples, n_channels)
    #     # waveform.shape is (n_samples, n_channels)
    #     features.append(np.einsum('ijk,jk->ki', pcs, waveform))
    features = np.einsum('ijk,ljk->lki', pcs, x)
    # features = np.stack(features, axis=0)
    assert features.ndim == 3
    return features


def compute_features(waveforms):
    assert waveforms.ndim == 3
    nspk, nsmp, nc = waveforms.shape
    pcs = _compute_pcs(waveforms, 3)
    assert pcs.ndim == 3
    features = _project_pcs(waveforms, pcs)
    assert features.ndim == 3
    assert features.shape == (nspk, nc, 3)
    return features


#------------------------------------------------------------------------------
# I/O util functions
#------------------------------------------------------------------------------

def _find_first_existing_path(*paths, multiple_ok=True):
    out = []
    for path in paths:
        path = Path(path)
        if path.exists():
            out.append(path)
    if len(out) >= 2 and not multiple_ok:  # pragma: no cover
        raise IOError("Multiple conflicting files exist: %s." % ', '.join((out, path)))
    elif len(out) >= 1:
        return out[0]
    else:
        return None


def _close_memmap(name, obj):
    """Close a memmap array or a list of memmap arrays."""
    if isinstance(obj, np.memmap):
        logger.debug("Close memmap array %s.", name)
        obj._mmap.close()
    elif getattr(obj, 'arrs', None) is not None:  # pragma: no cover
        # Support ConcatenatedArrays.
        # NOTE: no longer used since EphysTraces
        _close_memmap('%s.arrs' % name, obj.arrs)
    elif isinstance(obj, (list, tuple)):
        [_close_memmap('%s[]' % name, item) for item in obj]
    elif isinstance(obj, dict):
        [_close_memmap('%s.%s' % (name, n), item) for n, item in obj.items()]


#------------------------------------------------------------------------------
# Template model
#------------------------------------------------------------------------------

# Special spike_*.npy files that should not be considered as "spike attributes".
SKIP_SPIKE_ATTRS = ('clusters', 'templates', 'samples', 'times', 'times_reordered', 'amplitudes')


class TemplateModel(object):
    """Object holding all data of a KiloSort/phy dataset.

    Constructor
    -----------

    dir_path : str or Path
        Path to the dataset directory
    dat_path : str, Path, or list
        Path to the raw data files.
    dtype : NumPy dtype
        Data type of the raw data file
    offset : int
        Header offset of the binary file
    n_channels_dat : int
        Number of channels in the dat file
    sample_rate : float
        Sampling rate of the data file.

    """

    """Number of closest channels used for templates."""
    n_closest_channels = 12

    """Fraction of the peak amplitude required by the closest channels to be kept as best
    channels."""
    amplitude_threshold = 0

    def __init__(self, **kwargs):
        # Default empty values.
        self.dat_path = []
        self.sample_rate = None
        self.n_channels_dat = None

        self.__dict__.update(kwargs)

        # Set dir_path.
        self.dir_path = Path(self.dir_path).resolve()
        assert isinstance(self.dir_path, Path)
        assert self.dir_path.exists()

        # Set dat_path.
        if not self.dat_path:  # pragma: no cover
            self.dat_path = []
        elif not isinstance(self.dat_path, (list, tuple)):
            self.dat_path = [self.dat_path]
        assert isinstance(self.dat_path, (list, tuple))
        self.dat_path = [Path(p).resolve() if not Path(p).is_symlink() else p for p in self.dat_path]

        self.dtype = getattr(self, 'dtype', np.int16)
        if not self.sample_rate:  # pragma: no cover
            logger.warning("No sample rate was given! Defaulting to 1 Hz.")
        self.sample_rate = float(self.sample_rate or 1.)
        assert self.sample_rate > 0
        self.offset = getattr(self, 'offset', 0)

        self._load_data()

    #--------------------------------------------------------------------------
    # Internal loading methods
    #--------------------------------------------------------------------------

    def _load_data(self):
        """Load all data."""
        # Spikes
        self.spike_samples, self.spike_times = self._load_spike_samples()
        ns, = self.n_spikes, = self.spike_times.shape

        # Make sure the spike times are increasing.
        if not np.all(np.diff(self.spike_times) >= 0):
            raise ValueError("The spike times must be increasing.")

        # Spike amplitudes.
        self.amplitudes = self._load_amplitudes()
        if self.amplitudes is not None:
            assert self.amplitudes.shape == (ns,)

        # Spike templates.
        self.spike_templates = self._load_spike_templates()
        assert self.spike_templates.shape == (ns,)

        # Unique template ids.
        self.template_ids = np.unique(self.spike_templates)

        # Spike clusters.
        self.spike_clusters = self._load_spike_clusters()
        assert self.spike_clusters.shape == (ns,)

        # Unique cluster ids.
        self.cluster_ids = np.unique(self.spike_clusters)

        # Spike reordering.
        self.spike_times_reordered = self._load_spike_reorder()
        if self.spike_times_reordered is not None:
            assert self.spike_times_reordered.shape == (ns,)

        # Channels.
        self.channel_mapping = self._load_channel_map()
        self.n_channels = nc = self.channel_mapping.shape[0]
        if self.n_channels_dat:
            assert np.all(self.channel_mapping <= self.n_channels_dat - 1)

        # Channel positions.
        self.channel_positions = self._load_channel_positions()
        assert self.channel_positions.shape == (nc, 2)
        if not _all_positions_distinct(self.channel_positions):  # pragma: no cover
            logger.error(
                "Some channels are on the same position, please check the channel positions file.")
            self.channel_positions = linear_positions(nc)

        # Channel shanks.
        self.channel_shanks = self._load_channel_shanks()
        assert self.channel_shanks.shape == (nc,)

        # Channel probes.
        self.channel_probes = self._load_channel_probes()
        assert self.channel_probes.shape == (nc,)
        self.probes = np.unique(self.channel_probes)
        self.n_probes = len(self.probes)

        # Templates.
        self.sparse_templates = self._load_templates()
        if self.sparse_templates is not None:
            self.n_templates, self.n_samples_waveforms, self.n_channels_loc = \
                self.sparse_templates.data.shape
            if self.sparse_templates.cols is not None:
                assert self.sparse_templates.cols.shape == (self.n_templates, self.n_channels_loc)
        else:  # pragma: no cover
            self.n_templates = self.spike_templates.max() + 1
            self.n_samples_waveforms = 0
            self.n_channels_loc = 0

        # Clusters waveforms
        if not np.all(self.spike_clusters == self.spike_templates) and \
                self.sparse_templates.cols is None:
            self.merge_map, self.nan_idx = self.get_merge_map()
            self.sparse_clusters = self.cluster_waveforms()
            self.n_clusters = self.spike_clusters.max() + 1
        else:
            self.merge_map = {}
            self.nan_idx = []
            self.sparse_clusters = self.sparse_templates
            self.n_clusters = self.spike_templates.max() + 1

        # Spike waveforms (optional, otherwise fetched from raw data as needed).
        self.spike_waveforms = self._load_spike_waveforms()

        # Whitening.
        try:
            self.wm = self._load_wm()
        except IOError:
            logger.debug("Whitening matrix file not found.")
            self.wm = np.eye(nc)
        assert self.wm.shape == (nc, nc)
        try:
            self.wmi = self._load_wmi()
        except IOError:
            logger.debug("Whitening matrix inverse file not found, computing it.")
            self.wmi = self._compute_wmi(self.wm)
        assert self.wmi.shape == (nc, nc)

        # Similar templates.
        self.similar_templates = self._load_similar_templates()
        assert self.similar_templates.shape == (self.n_templates, self.n_templates)

        # Traces and duration.
        self.traces = self._load_traces(self.channel_mapping)
        if self.traces is not None:
            self.duration = self.traces.duration
        else:
            self.duration = self.spike_times[-1]
        if self.spike_times[-1] > self.duration:  # pragma: no cover
            logger.warning(
                "There are %d/%d spikes after the end of the recording.",
                np.sum(self.spike_times > self.duration), self.n_spikes)

        # Features.
        self.sparse_features = self._load_features()
        self.features = self.sparse_features.data if self.sparse_features else None
        if self.sparse_features is not None:
            self.n_features_per_channel = self.sparse_features.data.shape[2]

        # Template features.
        self.sparse_template_features = self._load_template_features()
        self.template_features = (
            self.sparse_template_features.data if self.sparse_template_features else None)

        # Spike attributes.
        self.spike_attributes = self._load_spike_attributes()

        # Metadata.
        self.metadata = self._load_metadata()

    def _find_path(self, *names, multiple_ok=True, mandatory=True):
        full_paths = list(l[0] for l in [list(self.dir_path.glob(name)) for name in names] if l)
        path = _find_first_existing_path(*full_paths, multiple_ok=multiple_ok)
        if mandatory and not path:
            raise IOError(
                "None of these files could be found in %s: %s." %
                (self.dir_path, ', '.join(names)))
        return path

    def _read_array(self, path, mmap_mode=None):
        if not path:
            raise IOError()
        return read_array(path, mmap_mode=mmap_mode).squeeze()

    def _write_array(self, path, arr):
        return write_array(path, arr)

    def _load_metadata(self):
        """Load cluster metadata from all CSV/TSV files in the data directory."""
        # Files to exclude.
        excluded_names = ('cluster_info',)
        # Get all CSV/TSV files in the directory.
        files = list(self.dir_path.glob('*.csv'))
        files.extend(self.dir_path.glob('*.tsv'))
        metadata = {}
        for filename in files:
            if filename.stem in excluded_names:
                continue
            logger.debug("Load `%s`.", filename.name)
            try:
                for field, data in load_metadata(filename).items():
                    metadata[field] = data
            except Exception as e:
                logger.warning("Error when reading %s: %s.", filename.name, str(e))
                continue
        return metadata

    #--------------------------------------------------------------------------
    # Specific loading methods
    #--------------------------------------------------------------------------

    def _load_spike_attributes(self):
        """Load all spike_*.npy files, called spike attributes."""
        files = list(self.dir_path.glob('spike_*.npy'))
        spike_attributes = Bunch()
        for filename in files:
            # The part after spike_***
            n = filename.stem[6:]
            # Skip known files.
            if n in SKIP_SPIKE_ATTRS:
                continue
            try:
                arr = self._read_array(filename)
                assert arr.shape[0] == self.n_spikes
                logger.debug("Load %s.", filename.name)
            except (IOError, AssertionError) as e:
                logger.warning("Unable to open %s: %s.", filename.name, e)
                continue
            spike_attributes[n] = arr
        return spike_attributes

    def _load_channel_map(self):
        path = self._find_path('channel_map.npy', 'channels.rawInd*.npy')
        out = self._read_array(path)
        out = np.atleast_1d(out)
        assert out.ndim == 1
        assert out.dtype in (np.uint32, np.int32, np.int64)
        return out

    def _load_channel_positions(self):
        path = self._find_path('channel_positions.npy', 'channels.localCoordinates*.npy')
        out = self._read_array(path)
        out = np.atleast_2d(out)
        assert out.ndim == 2
        return out

    def _load_channel_probes(self):
        try:
            path = self._find_path('channel_probe.npy', 'channels.probes*.npy')
            out = self._read_array(path)
            out = np.atleast_1d(out)
            assert out.ndim == 1
            return out
        except IOError:
            return np.zeros(self.n_channels, dtype=np.int32)

    def _load_channel_shanks(self):
        try:
            path = self._find_path('channel_shanks.npy', 'channels.shanks*.npy')
            out = self._read_array(path).reshape((-1,))
            assert out.ndim == 1
            return out
        except IOError:
            logger.debug("No channel shank file found.")
            return np.zeros(self.n_channels, dtype=np.int32)

    def _load_traces(self, channel_map=None):
        if not self.dat_path:
            if os.environ.get('PHY_VIRTUAL_RAW_DATA', None):  # pragma: no cover
                n_samples = int((self.spike_times[-1] + 1) * self.sample_rate)
                return RandomEphysReader(n_samples, len(channel_map), sample_rate=self.sample_rate)
            return
        n = self.n_channels_dat
        # self.dat_path could be any object accepted by get_ephys_reader().
        traces = get_ephys_reader(
            self.dat_path, n_channels_dat=n, dtype=self.dtype, offset=self.offset,
            sample_rate=self.sample_rate)
        if traces is not None:
            traces = traces[:, channel_map]  # lazy permutation on the channel axis
        return traces

    def _load_amplitudes(self):
        try:
            out = self._read_array(self._find_path('amplitudes.npy', 'spikes.amps*.npy'))
            assert out.ndim == 1
            return out
        except IOError:
            logger.debug("No amplitude file found.")
            return

    def _load_spike_templates(self):
        path = self._find_path('spike_templates.npy', 'spikes.templates*.npy')
        out = self._read_array(path)
        if out.dtype in (np.float32, np.float64):  # pragma: no cover
            out = out.astype(np.int32)
        uc = np.unique(out)
        if np.max(uc) - np.min(uc) + 1 != uc.size:
            logger.warning(
                "Unreferenced clusters found in templates (generally not a problem)")
        assert out.dtype in (np.uint16, np.uint32, np.int32, np.int64)
        assert out.ndim == 1
        return out

    def _load_spike_clusters(self):
        path = self._find_path(
            'spike_clusters.npy', 'spikes.clusters*.npy', multiple_ok=False, mandatory=False)
        if path is None:
            # Create spike_clusters file if it doesn't exist.
            tmp_path = self._find_path('spike_templates.npy', 'spikes.clusters*.npy')
            path = self.dir_path / 'spike_clusters.npy'
            logger.debug("Copying from %s to %s.", tmp_path, path)
            shutil.copy(tmp_path, path)
        assert path.exists()
        logger.debug("Loading spike clusters.")
        # NOTE: we make a copy in memory so that we can update this array
        # during manual clustering.
        out = self._read_array(path).astype(np.int32)
        uc = np.unique(out)
        if np.max(uc) - np.min(uc) + 1 != uc.size:
            logger.warning(
                "Unreferenced clusters found in spike_clusters (generally not a problem)")
        assert out.ndim == 1
        return out

    def _load_spike_reorder(self):
        """Load spike_times_reordered.npy, a 1D array with alternative spike times, in number of
        samples (not in seconds)."""
        path = self.dir_path / 'spike_times_reordered.npy'
        if path.exists():
            logger.debug("Loading spike times reordered.")
            samples = self._read_array(path).squeeze()
            times = samples / self.sample_rate
            assert times.shape == (self.n_spikes,)
            return times

    def _load_spike_samples(self):
        # WARNING: "spike_times.npy" is in units of samples. Need to
        # divide by the sampling rate to get spike times in seconds.
        path = self.dir_path / 'spike_times.npy'
        if path.exists():
            # WARNING: spikes_times.npy is in samples !
            samples = self._read_array(path)
            times = samples / self.sample_rate
        else:
            # WARNING: spikes.times.npy is in seconds, not samples !
            times_path = self._find_path('spikes.times*.npy', mandatory=False)
            times = self._read_array(times_path)
            samples_path = self._find_path('spikes.samples*.npy', mandatory=False)
            if samples_path:
                samples = self._read_array(samples_path)
            else:
                logger.info("Loading spikes.times.npy in seconds, converting to samples.")
                samples = np.round(times * self.sample_rate).astype(np.uint64)
        assert samples.ndim == times.ndim == 1
        return samples, times

    def _load_spike_waveforms(self):  # pragma: no cover
        path = self.dir_path / '_phy_spikes_subset.waveforms.npy'
        path_channels = self.dir_path / '_phy_spikes_subset.channels.npy'
        path_spikes = self.dir_path / '_phy_spikes_subset.spikes.npy'
        if not path.exists() or not path_channels.exists() or not path_spikes.exists():
            logger.warning(
                "Skipping spike waveforms that do not exist, they will be extracted "
                "on the fly from the raw data as needed.")
            return
        logger.debug("Loading spikes subset waveforms to avoid fetching waveforms from raw data.")
        try:
            return Bunch(
                waveforms=self._read_array(path, mmap_mode='r'),
                spike_channels=self._read_array(path_channels),
                spike_ids=self._read_array(path_spikes),
            )
        except Exception as e:
            logger.warning("Could not load spike waveforms: %s.", e)
            return

    def _load_similar_templates(self):
        try:
            out = self._read_array(self._find_path('similar_templates.npy'))
            out = np.atleast_2d(out)
            assert out.ndim == 2
            return out
        except IOError:
            return np.zeros((self.n_templates, self.n_templates))

    def _load_templates(self):
        logger.debug("Loading templates.")

        # Sparse structure: regular array with col indices.
        try:
            path = self._find_path(
                'templates.npy', 'templates.waveforms.npy', 'templates.waveforms.*.npy')
            data = self._read_array(path, mmap_mode='r+')
            data = np.atleast_3d(data)
            assert data.ndim == 3
            assert data.dtype in (np.float32, np.float64)
            # WARNING: this will load the full array in memory, might cause memory problems
            empty_templates = np.all(np.all(np.isnan(data), axis=1), axis=1)
            data[empty_templates, ...] = 0
            n_templates, n_samples, n_channels_loc = data.shape
        except IOError:
            return

        try:
            # WARNING: KS2 saves templates_ind.npy (with an s), and note template_ind.npy,
            # so that file is not taken into account here.
            # That means templates.npy is considered as a dense array.
            # Proper fix would be to save templates.npy as a true sparse array, with proper
            # template_ind.npy (without an s).
            path = self._find_path('template_ind.npy', 'templates.waveformsChannels*.npy')
            cols = self._read_array(path)
            if cols.ndim != 2:  # pragma: no cover
                cols = np.atleast_2d(cols).T
            assert cols.ndim == 2
            logger.debug("Templates are sparse.")

            assert cols.shape == (n_templates, n_channels_loc)
        except IOError:
            logger.debug("Templates are dense.")
            cols = None

        return Bunch(data=data, cols=cols)

    def _load_wm(self):
        logger.debug("Loading the whitening matrix.")
        out = self._read_array(self._find_path('whitening_mat.npy'))
        out = np.atleast_2d(out)
        assert out.ndim == 2
        return out

    def _load_wmi(self):  # pragma: no cover
        logger.debug("Loading the inverse of the whitening matrix.")
        out = self._read_array(self._find_path('whitening_mat_inv.npy'))
        out = np.atleast_2d(out)
        assert out.ndim == 2
        return out

    def _compute_wmi(self, wm):
        logger.debug("Inversing the whitening matrix %s.", wm.shape)
        try:
            wmi = np.linalg.inv(wm)
        except np.linalg.LinAlgError as e:  # pragma: no cover
            raise ValueError(
                "Error when inverting the whitening matrix: %s.", e)
        self._write_array(self.dir_path / 'whitening_mat_inv.npy', wmi)
        return wmi

    def _unwhiten(self, x, channel_ids=None):
        mat = self.wmi
        if channel_ids is not None:
            mat = mat[np.ix_(channel_ids, channel_ids)]
            assert mat.shape == (len(channel_ids),) * 2
        assert x.shape[1] == mat.shape[0]
        out = np.dot(x, mat) * getattr(self, 'template_scaling', 1.0)
        return np.ascontiguousarray(out)

    def _load_features(self):

        # Sparse structure: regular array with row and col indices.
        try:
            logger.debug("Loading features.")
            data = self._read_array(
                self._find_path('pc_features.npy'), mmap_mode='r')
            if data.ndim == 2:  # pragma: no cover
                # Deal with npcs = 1.
                data = data.reshape(data.shape + (1,))
            assert data.ndim == 3
            assert data.dtype in (np.float32, np.float64)
            data = data.transpose((0, 2, 1))
            n_spikes, n_channels_loc, n_pcs = data.shape
        except IOError:
            return

        try:
            cols = self._read_array(self._find_path('pc_feature_ind.npy'), mmap_mode='r')
            logger.debug("Features are sparse.")
            if cols.ndim == 1:  # pragma: no cover
                # Deal with npcs = 1.
                cols = cols.reshape(cols.shape + (1,))
            assert cols.ndim == 2
            assert cols.shape == (self.n_templates, n_channels_loc)
        except IOError:
            logger.debug("Features are dense.")
            cols = None

        try:
            rows = self._read_array(self._find_path('pc_feature_spike_ids.npy'))
            assert rows.shape == (n_spikes,)
        except IOError:
            rows = None

        return Bunch(data=data, cols=cols, rows=rows)

    def _load_template_features(self):

        # Sparse structure: regular array with row and col indices.
        try:
            logger.debug("Loading template features.")
            data = self._read_array(self._find_path('template_features.npy'), mmap_mode='r')
            assert data.dtype in (np.float32, np.float64)
            assert data.ndim == 2
            n_spikes, n_channels_loc = data.shape
        except IOError:
            return

        try:
            cols = self._read_array(self._find_path('template_feature_ind.npy'))
            logger.debug("Template features are sparse.")
            assert cols.shape == (self.n_templates, n_channels_loc)
        except IOError:
            cols = None
            logger.debug("Template features are dense.")

        try:
            rows = self._read_array(self._find_path('template_feature_spike_ids.npy'))
            assert rows.shape == (n_spikes,)
        except IOError:
            rows = None

        return Bunch(data=data, cols=cols, rows=rows)

    #--------------------------------------------------------------------------
    # Internal data access methods
    #--------------------------------------------------------------------------

    def _find_best_channels(self, template, amplitude_threshold=None):
        """Find the best channels for a given template."""
        # Compute the template amplitude on each channel.
        assert template.ndim == 2  # shape: (n_samples, n_channels)
        amplitude = template.max(axis=0) - template.min(axis=0)
        assert not np.all(np.isnan(amplitude)), "Template is all NaN!"
        assert amplitude.ndim == 1  # shape: (n_channels,)
        # Find the peak channel.
        best_channel = np.argmax(amplitude)
        max_amp = amplitude[best_channel]
        # Find the channels X% peak.
        amplitude_threshold = (
            amplitude_threshold if amplitude_threshold is not None else self.amplitude_threshold)
        peak_channels = np.nonzero(amplitude >= amplitude_threshold * max_amp)[0]
        # Find N closest channels.
        close_channels = get_closest_channels(
            self.channel_positions, best_channel, self.n_closest_channels)
        assert best_channel in close_channels
        # Restrict to the channels belonging to the best channel's shank.
        if self.channel_shanks is not None:
            shank = self.channel_shanks[best_channel]  # shank of best channel
            channels_on_shank = np.nonzero(self.channel_shanks == shank)[0]
            close_channels = np.intersect1d(close_channels, channels_on_shank)
        # Keep the intersection.
        channel_ids = np.intersect1d(peak_channels, close_channels)
        # Order the channels by decreasing amplitude.
        order = np.argsort(amplitude[channel_ids])[::-1]
        channel_ids = channel_ids[order]
        amplitude = amplitude[order]
        assert best_channel in channel_ids
        assert amplitude.shape == (len(channel_ids),)
        return channel_ids, amplitude, best_channel

    def _template_n_channels(self, template_id, n_channels):
        """Return the n best channels for a given template, filling with -1s if there isn't
        enough best channels for that template."""
        assert n_channels > 0
        if template_id not in self.template_ids:
            return [-1] * n_channels
        template = self.get_template(template_id)
        channel_ids = list(template.channel_ids[:n_channels])
        if len(channel_ids) < n_channels:
            channel_ids += [-1] * (n_channels - len(channel_ids))
        return channel_ids

    def _get_template_dense(self, template_id, channel_ids=None, amplitude_threshold=None,
                            unwhiten=True):
        """Return data for one template."""
        if not self.sparse_templates:
            return
        template_w = self.sparse_templates.data[template_id, ...]
        template = self._unwhiten(template_w).astype(np.float32) if unwhiten else template_w
        assert template.ndim == 2
        channel_ids_, amplitude, best_channel = self._find_best_channels(
            template, amplitude_threshold=amplitude_threshold)
        channel_ids = channel_ids if channel_ids is not None else channel_ids_
        template = template[:, channel_ids]
        assert template.ndim == 2
        assert template.shape[1] == channel_ids.shape[0]
        return Bunch(
            template=template,
            amplitude=amplitude,
            best_channel=best_channel,
            channel_ids=channel_ids,
        )

    def _get_template_sparse(self, template_id, unwhiten=True):
        data, cols = self.sparse_templates.data, self.sparse_templates.cols
        assert cols is not None
        template_w, channel_ids = data[template_id], cols[template_id]

        # KS2 HACK: dense templates may have been saved as sparse arrays (with all channels),
        # we need to remove channels with no signal.

        # template_w is (n_samples, n_channels)
        template_max = np.abs(template_w).max(axis=0)  # n_channels
        has_signal = template_max > template_max.max() * 1e-6
        channel_ids = channel_ids[has_signal]
        template_w = template_w[:, has_signal]

        # Remove unused channels = -1.
        used = channel_ids != -1
        template_w = template_w[:, used]
        channel_ids = channel_ids[used]
        channel_ids = channel_ids.astype(np.uint32)

        # Unwhiten.
        template = self._unwhiten(template_w, channel_ids=channel_ids) if unwhiten else template_w
        template = template.astype(np.float32)
        assert template.ndim == 2
        assert template.shape[1] == len(channel_ids)
        # Compute the amplitude and the channel with max amplitude.
        amplitude = template.max(axis=0) - template.min(axis=0)
        best_channel = channel_ids[np.argmax(amplitude)]
        # NOTE: it is expected that the channel_ids are reordered by decreasing amplitude.
        # To each column of the template array corresponds the channel id given by channel_ids.
        channels_reordered = np.argsort(amplitude)[::-1]
        out = Bunch(
            template=template[..., channels_reordered],
            amplitude=amplitude,
            best_channel=best_channel,
            channel_ids=channel_ids[channels_reordered],
        )
        return out

    def get_merge_map(self):
        """"Gets the maps of merges and splits between spikes.clusters and spikes.templates"""
        inverse_mapping_dict = {key: [] for key in range(np.max(self.spike_clusters) + 1)}
        for temp in np.unique(self.spike_templates):
            idx = np.where(self.spike_templates == temp)[0]
            new_idx = self.spike_clusters[idx]
            mapping = np.unique(new_idx)
            for n in mapping:
                inverse_mapping_dict[n].append(temp)

        nan_idx = np.array([idx for idx, val in inverse_mapping_dict.items() if len(val) == 0])

        return inverse_mapping_dict, nan_idx

    #--------------------------------------------------------------------------
    # Data access methods
    #--------------------------------------------------------------------------

    def get_template(self, template_id, channel_ids=None, amplitude_threshold=None, unwhiten=True):
        """Get data about a template."""
        if self.sparse_templates and self.sparse_templates.cols is not None:
            return self._get_template_sparse(template_id, unwhiten=unwhiten)
        else:
            return self._get_template_dense(
                template_id, channel_ids=channel_ids, amplitude_threshold=amplitude_threshold,
                unwhiten=unwhiten)

    def get_waveforms(self, spike_ids, channel_ids=None):
        """Return spike waveforms on specified channels."""
        if self.traces is None and self.spike_waveforms is None:
            return
        # Create the output array.
        nsw = self.n_samples_waveforms
        channel_ids = np.arange(self.n_channels) if channel_ids is None else channel_ids

        if self.spike_waveforms is not None:
            # Load from precomputed spikes.
            try:
                return get_spike_waveforms(
                spike_ids, channel_ids, spike_waveforms=self.spike_waveforms,
                n_samples_waveforms=nsw)
            except AssertionError:
              logger.warning(
                  "Error when loading waveforms from precomputed waveforms, trying to load the raw data.")
              spike_samples = self.spike_samples[spike_ids]
              return extract_waveforms(
                  self.traces, spike_samples, channel_ids, n_samples_waveforms=nsw)
        else:
            # Or load directly from raw data (slower).
            spike_samples = self.spike_samples[spike_ids]
            return extract_waveforms(
                self.traces, spike_samples, channel_ids, n_samples_waveforms=nsw)

    def get_features(self, spike_ids, channel_ids):
        """Return sparse features for given spikes."""
        sf = self.sparse_features
        if sf is None and self.spike_waveforms is not None:
            ns = len(spike_ids)
            nc = len(channel_ids)
            n_pcs = 3
            features = np.zeros((ns, nc, n_pcs), dtype=np.float32)
            spike_ids_exist = np.intersect1d(spike_ids, self.spike_waveforms.spike_ids)
            # Compute PCs from the waveforms for the spikes that are in spike_waveforms.spike_ids.
            waveforms = self.get_waveforms(spike_ids_exist, channel_ids)
            features_existing = compute_features(waveforms)
            assert features.shape[1:] == (nc, n_pcs)
            # Now we need to integrate the computed features into the output array, knowing
            # that some spikes may be missing if there were requested here in spike_ids, but
            # were absent in spike_waveforms.spike_ids.
            ind = _index_of(spike_ids_exist, spike_ids)
            features[ind, ...] = features_existing
            return features
        elif sf is None:
            return
        _, n_channels_loc, n_pcs = sf.data.shape
        ns = len(spike_ids)
        nc = len(channel_ids)

        # Initialize the output array.
        features = np.empty((ns, n_channels_loc, n_pcs))
        features[:] = np.nan

        if sf.rows is not None:
            s = np.intersect1d(spike_ids, sf.rows)
            # Relative indices of the spikes in the self.features_spike_ids
            # array, necessary to load features from all_features which only
            # contains the subset of the spikes.
            rows = _index_of(s, sf.rows)
            # Relative indices of the non-null rows in the output features
            # array.
            rows_out = _index_of(s, spike_ids)
        else:
            rows = spike_ids
            rows_out = slice(None, None, None)
        features[rows_out, ...] = sf.data[rows]

        if sf.cols is not None:
            assert sf.cols.shape[1] == n_channels_loc
            cols = sf.cols[self.spike_templates[spike_ids]]
        else:
            cols = np.tile(np.arange(n_channels_loc), (ns, 1))
        features = from_sparse(features, cols, channel_ids)

        assert features.shape == (ns, nc, n_pcs)
        return features

    def get_template_features(self, spike_ids):
        """Return sparse template features for given spikes."""
        tf = self.sparse_template_features
        if tf is None:
            return
        _, n_templates_loc = tf.data.shape
        ns = len(spike_ids)

        if tf.rows is not None:
            spike_ids = np.intersect1d(spike_ids, tf.rows)
            # Relative indices of the spikes in the self.features_spike_ids
            # array, necessary to load features from all_features which only
            # contains the subset of the spikes.
            rows = _index_of(spike_ids, tf.rows)
        else:
            rows = spike_ids
        template_features = tf.data[rows]

        if tf.cols is not None:
            assert tf.cols.shape[1] == n_templates_loc
            cols = tf.cols[self.spike_templates[spike_ids]]
        else:
            cols = np.tile(np.arange(n_templates_loc), (len(spike_ids), 1))
        template_features = from_sparse(template_features, cols, np.arange(self.n_templates))

        assert template_features.shape[0] == ns
        return template_features

    def get_depths(self):
        """Compute spike depths based on spike pc features and probe depths."""
        # compute the depth as the weighted sum of coordinates
        # if PC features are provided, compute the depth as the weighted sum of coordinates
        nbatch = 50000
        c = 0
        spikes_depths = np.zeros_like(self.spike_times) * np.nan
        nspi = spikes_depths.shape[0]
        if self.sparse_features is None or self.sparse_features.data.shape[0] != self.n_spikes:
            return None
        while True:
            ispi = np.arange(c, min(c + nbatch, nspi))
            # take only first component
            features = self.sparse_features.data[ispi, :, 0]
            features = np.maximum(features, 0) ** 2  # takes only positive values into account
            ichannels = self.sparse_features.cols[self.spike_templates[ispi]].astype(np.uint32)
            # features = np.square(self.sparse_features.data[ispi, :, 0])
            # ichannels = self.sparse_features.cols[self.spike_templates[ispi]].astype(np.int64)
            ypos = self.channel_positions[ichannels, 1]
            with np.errstate(divide='ignore'):
                spikes_depths[ispi] = (np.sum(np.transpose(ypos * features) /
                                              np.sum(features, axis=1), axis=0))
            c += nbatch
            if c >= nspi:
                break
        return spikes_depths

    def get_amplitudes_true(self, sample2unit=1., use='templates'):
        """Convert spike amplitude values to input amplitudes units
         via scaling by unwhitened template waveform.
         :param sample2unit float: factor to convert the raw data to a physical unit (defaults 1.)
         :returns: spike_amplitudes_volts: np.array [nspikes] spike amplitudes in raw data units
         :returns: templates_volts: np.array[ntemplates, nsamples, nchannels]: templates
         in raw data units
         :returns: template_amps_volts: np.array[ntemplates]: average templates amplitudes
          in raw data units
         To scale the template for template matching,
         raw_data_volts = templates_volts * spike_amplitudes_volts / template_amps_volts
         """
        # spike_amp = ks2_spike_amps * maxmin(inv_whitening(ks2_template_amps))
        # to rescale the template,

        if use == 'clusters':
            sparse = self.sparse_clusters
            spikes = self.spike_clusters
            n_wav = self.n_clusters
        else:
            sparse = self.sparse_templates
            spikes = self.spike_templates
            n_wav = self.n_templates

        # unwhiten template waveforms on their channels of max amplitude
        if sparse.cols:
            raise NotImplementedError
        # apply the inverse whitening matrix to the template
        templates_wfs = np.zeros_like(sparse.data)  # nt, ns, nc
        for n in np.arange(n_wav):
            templates_wfs[n, :, :] = np.matmul(sparse.data[n, :, :], self.wmi)

        # The amplitude on each channel is the positive peak minus the negative
        templates_ch_amps = np.max(templates_wfs, axis=1) - np.min(templates_wfs, axis=1)

        # The template arbitrary unit amplitude is the amplitude of its largest channel
        # (but see below for true tempAmps)
        templates_amps_au = np.max(templates_ch_amps, axis=1)
        spike_amps = templates_amps_au[spikes] * self.amplitudes

        with np.errstate(divide='ignore', invalid='ignore'):
            # take the average spike amplitude per template
            templates_amps_v = (np.bincount(spikes, weights=spike_amps) /
                                np.bincount(spikes))
            # scale back the template according to the spikes units
            templates_physical_unit = templates_wfs * (templates_amps_v / templates_amps_au
                                                       )[:, np.newaxis, np.newaxis]

        return (spike_amps * sample2unit,
                templates_physical_unit * sample2unit,
                templates_amps_v * sample2unit)

    #--------------------------------------------------------------------------
    # Internal helper methods for public high-level methods
    #--------------------------------------------------------------------------

    def _get_template_from_spikes(self, spike_ids):
        """Get the main template from a set of spikes."""
        # We get the original template ids for the spikes.
        st = self.spike_templates[spike_ids]
        # We find the template with the largest number of spikes from the spike selection.
        template_ids, counts = np.unique(st, return_counts=True)
        ind = np.argmax(counts)
        template_id = template_ids[ind]
        # We load the template.
        template = self.get_template(template_id)
        # We return the template.
        return template

    #--------------------------------------------------------------------------
    # Public high-level methods
    #--------------------------------------------------------------------------

    def describe(self):
        """Display basic information about the dataset."""
        def _print(name, value):
            print("{0: <24}{1}".format(name, value))

        _print('Data files', ', '.join(map(str, self.dat_path)))
        _print('Directory', self.dir_path)
        _print('Duration', '{:.1f}s'.format(self.duration))
        _print('Sample rate', '{:.1f} kHz'.format(self.sample_rate / 1000.))
        _print('Data type', self.dtype)
        _print('# of channels', self.n_channels)
        _print('# of channels (raw)', self.n_channels_dat)
        _print('# of templates', self.n_templates)
        _print('# of spikes', "{:,}".format(self.n_spikes))

    def get_template_counts(self, cluster_id):
        """Return a histogram of the number of spikes in each template for a given cluster."""
        spike_ids = self.get_cluster_spikes(cluster_id)
        st = self.spike_templates[spike_ids]
        return np.bincount(st, minlength=self.n_templates)

    def get_template_spikes(self, template_id):
        """Return the spike ids that belong to a given template."""
        return _spikes_in_clusters(self.spike_templates, [template_id])

    def get_cluster_spikes(self, cluster_id):
        """Return the spike ids that belong to a given template."""
        return _spikes_in_clusters(self.spike_clusters, [cluster_id])

    def get_template_channels(self, template_id):
        """Return the most relevant channels of a template."""
        template = self.get_template(template_id)
        return template.channel_ids

    def get_cluster_channels(self, cluster_id):
        """Return the most relevant channels of a cluster."""
        spike_ids = self.get_cluster_spikes(cluster_id)
        return self._get_template_from_spikes(spike_ids).channel_ids

    def get_template_waveforms(self, template_id):
        """Return the waveforms of a template on the most relevant channels."""
        template = self.get_template(template_id)
        return template.template if template else None

    def get_cluster_mean_waveforms(self, cluster_id, unwhiten=True):
        """Return the mean template waveforms of a cluster, as a weighted average of the
        template waveforms from which the cluster originates from."""
        count = self.get_template_counts(cluster_id)
        best_template = np.argmax(count)
        template_ids = np.nonzero(count)[0]
        count = count[template_ids]
        # Get local channels of the best template for the given cluster.
        template = self.get_template(best_template, unwhiten=unwhiten)
        channel_ids = template.channel_ids
        # Get all templates from which this cluster stems from.
        templates = [self.get_template(template_id, unwhiten=unwhiten)
                     for template_id in template_ids]
        # Construct the waveforms array.
        ns = self.n_samples_waveforms
        data = np.zeros((len(template_ids), ns, self.n_channels))
        for i, b in enumerate(templates):
            data[i][:, b.channel_ids] = b.template
        waveforms = data[..., channel_ids]
        assert waveforms.shape == (len(template_ids), ns, len(channel_ids))
        mean_waveforms = np.average(waveforms, axis=0, weights=count)
        assert mean_waveforms.shape == (ns, len(channel_ids))
        return Bunch(mean_waveforms=mean_waveforms, channel_ids=channel_ids)

    def get_template_spike_waveforms(self, template_id):
        """Return all spike waveforms of a template, on the most relevant channels."""
        spike_ids = self.get_template_spikes(template_id)
        channel_ids = self.get_template_channels(template_id)
        return self.get_waveforms(spike_ids, channel_ids)

    def get_cluster_spike_waveforms(self, cluster_id):
        """Return all spike waveforms of a cluster, on the most relevant channels."""
        spike_ids = self.get_cluster_spikes(cluster_id)
        channel_ids = self.get_cluster_channels(cluster_id)
        return self.get_waveforms(spike_ids, channel_ids)

    @property
    def templates_channels(self):
        """Returns a vector of peak channels for all templates waveforms"""
        return self._channels(self.sparse_templates)

    @property
    def clusters_channels(self):
        """Returns a vector of peak channels for all clusters waveforms"""
        channels = self._channels(self.sparse_clusters)
        return channels

    def _channels(self, sparse):
        """ Gets peak channels for each waveform"""
        tmp = sparse.data
        n_templates, n_samples, n_channels = tmp.shape
        if sparse.cols is None:
            template_peak_channels = np.argmax(tmp.max(axis=1) - tmp.min(axis=1), axis=1)
        else:
            # when the templates are sparse, the first channel is the highest amplitude channel
            template_peak_channels = sparse.cols[:, 0]
        assert template_peak_channels.shape == (n_templates,)
        return template_peak_channels

    @property
    def templates_probes(self):
        """Returns a vector of probe index for all templates"""
        return self.channel_probes[self.templates_channels]

    @property
    def templates_amplitudes(self):
        """Returns the average amplitude per cluster"""
        return self._amplitudes(self.spike_templates)

    @property
    def clusters_amplitudes(self):
        """Returns the average amplitude per cluster"""
        return self._amplitudes(self.spike_clusters)

    def _amplitudes(self, tmp):
        """ Compute average amplitude for spikes"""
        tid = np.unique(tmp)
        n = np.bincount(tmp)[tid]
        a = np.bincount(tmp, weights=self.amplitudes)[tid]
        n[np.isnan(n)] = 1
        return a / n

    @property
    def templates_waveforms_durations(self):
        """Returns a vector of waveform durations (ms) for all templates"""
        return self._waveform_durations(self.sparse_templates.data)

    @property
    def clusters_waveforms_durations(self):
        """Returns a vector of waveform durations (ms) for all clusters"""
        waveform_duration = self._waveform_durations(self.sparse_clusters.data)
        return waveform_duration

    def _waveform_durations(self, tmp):
        n_templates, n_samples, n_channels = tmp.shape
        # Compute the peak channels for each template.
        template_peak_channels = np.argmax(tmp.max(axis=1) - tmp.min(axis=1), axis=1)
        durations = tmp.argmax(axis=1) - tmp.argmin(axis=1)
        ind = np.ravel_multi_index((np.arange(0, n_templates), template_peak_channels),
                                   (n_templates, n_channels), mode='raise', order='C')
        return durations.flatten()[ind].astype(np.float64) / self.sample_rate * 1e3

    def cluster_waveforms(self):
        """
        Computes the cluster waveforms for split and merged clusters
        :return:
        """
        # Only non sparse implementation
        ns = self.n_samples_waveforms
        data = np.zeros((np.max(self.cluster_ids) + 1, ns, self.n_channels))
        for clust, val in self.merge_map.items():
            if len(val) > 1:
                mean_waveform = self.get_cluster_mean_waveforms(clust, unwhiten=False)
                data[clust, :, mean_waveform.channel_ids] = \
                    np.swapaxes(mean_waveform.mean_waveforms, 0, 1)
            elif len(val) == 1:
                data[clust, :, :] = self.sparse_templates.data[val[0], :, :]

        return Bunch(data=data, cols=None)

    #--------------------------------------------------------------------------
    # Saving methods
    #--------------------------------------------------------------------------

    def save_metadata(self, name, values):
        """Save a dictionary {cluster_id: value} with cluster metadata in
        a TSV file."""
        path = self.dir_path / ('cluster_%s.tsv' % name)
        logger.debug("Save cluster metadata to `%s`.", path)
        # Remove empty values.
        save_metadata(
            path, name, {c: v for c, v in values.items() if v is not None})

    def save_spike_clusters(self, spike_clusters):
        """Save the spike clusters."""
        path = self._find_path('spike_clusters.npy', 'spikes.clusters.npy', multiple_ok=False)
        logger.debug("Save spike clusters to `%s`.", path)
        np.save(path, spike_clusters)

    def save_spikes_subset_waveforms(self, max_n_spikes_per_template=None, max_n_channels=None,
                                     sample2unit=1.):
        if self.traces is None:
            logger.warning(
                "Spike waveforms could not be extracted as the raw data file is not available.")
            return

        n_chunks_kept = 20  # TODO: better choice
        nst = max_n_spikes_per_template
        nc = max_n_channels or self.n_closest_channels
        nc = max(nc, self.n_closest_channels)

        assert nst > 0
        assert nc > 0

        path = self.dir_path / '_phy_spikes_subset.waveforms.npy'
        path_spikes = self.dir_path / '_phy_spikes_subset.spikes.npy'
        path_channels = self.dir_path / '_phy_spikes_subset.channels.npy'

        # Subselection of spikes.
        spt = _spikes_per_cluster(self.spike_templates)
        template_ids = sorted(spt.keys())
        ss = SpikeSelector(
            get_spikes_per_cluster=lambda cl: spt.get(cl, np.array([], dtype=np.int64)),
            spike_times=self.spike_samples, chunk_bounds=self.traces.chunk_bounds,
            n_chunks_kept=n_chunks_kept)
        spike_ids = ss(max_n_spikes_per_template, template_ids, subset_chunks=True)

        # Save the spike ids.
        ns = len(spike_ids)
        logger.debug("Saving spike waveforms: %d spikes.", ns)
        np.save(path_spikes, spike_ids)

        # Save the spike channels.
        best_channels = np.vstack([
            self._template_n_channels(t, nc) for t in range(self.n_templates)]).astype(np.int32)
        assert best_channels.ndim == 2
        assert best_channels.shape[0] == self.n_templates
        spike_channels = best_channels[self.spike_templates[spike_ids], :]
        assert spike_channels.shape == (ns, nc)
        logger.debug("Saving spike waveforms: spike channels.")
        np.save(path_channels, spike_channels)

        # Extract waveforms from the raw data on a chunk by chunk basis.
        export_waveforms(
            path, self.traces, self.spike_samples[spike_ids], spike_channels,
            n_samples_waveforms=self.n_samples_waveforms, sample2unit=sample2unit)

        # Reload spike waveforms.
        self.spike_waveforms = self._load_spike_waveforms()

    def close(self):
        """Close all memmapped files."""
        for k, v in sorted(self.__dict__.items(), key=itemgetter(0)):
            _close_memmap(k, v)


def _make_abs_path(p, dir_path):
    p = Path(p)
    if not op.isabs(p):
        p = dir_path / p
    if not p.exists():
        logger.warning("File %s does not exist.", p)
    return p


def get_template_params(params_path):
    """Get a dictionary of parameters from a `params.py` file."""
    params_path = Path(params_path)

    params = read_python(params_path)
    params['dtype'] = np.dtype(params['dtype'])

    if 'dir_path' not in params:
        params['dir_path'] = params_path.parent
    params['dir_path'] = Path(params['dir_path'])
    assert params['dir_path'].is_dir()
    assert params['dir_path'].exists()

    if isinstance(params['dat_path'], str):
        params['dat_path'] = [params['dat_path']]
    params['dat_path'] = [_make_abs_path(_, params['dir_path']) for _ in params['dat_path']]
    return params


def load_model(params_path):
    """Return a TemplateModel instance from a path to a `params.py` file."""
    return TemplateModel(**get_template_params(params_path))



# -*- coding: utf-8 -*-

"""Test ALF dataset generation."""


#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import os
from pathlib import Path
import shutil
from pytest import fixture, raises

import numpy as np
import numpy.random as nr

from phylib.utils._misc import _write_tsv_simple
from ..alf import _FILE_RENAMES, _load, EphysAlfCreator
from ..model import TemplateModel


#------------------------------------------------------------------------------
# Fixture
#------------------------------------------------------------------------------

class Dataset(object):
    def __init__(self, tempdir):
        np.random.seed(42)
        self.tmp_dir = tempdir
        p = Path(self.tmp_dir)
        self.ns = 100
        self.nsamp = 25
        self.ncmax = 42
        self.nc = 10
        self.nt = 5
        self.ncd = 1000
        np.save(p / 'spike_times.npy', .01 * np.cumsum(nr.exponential(size=self.ns)))
        np.save(p / 'spike_clusters.npy', nr.randint(low=1, high=self.nt, size=self.ns))
        shutil.copy(p / 'spike_clusters.npy', p / 'spike_templates.npy')
        np.save(p / 'amplitudes.npy', nr.uniform(low=0.5, high=1.5, size=self.ns))
        np.save(p / 'channel_positions.npy', np.c_[np.arange(self.nc), np.zeros(self.nc)])
        np.save(p / 'templates.npy', np.random.normal(size=(self.nt, 50, self.nc)))
        np.save(p / 'similar_templates.npy', np.tile(np.arange(self.nt), (self.nt, 1)))
        np.save(p / 'channel_map.npy', np.c_[np.arange(self.nc)])
        np.save(p / 'channel_probe.npy', np.zeros(self.nc))
        np.save(p / 'whitening_mat.npy', np.eye(self.nc, self.nc))
        np.save(p / '_phy_spikes_subset.channels.npy', np.zeros([self.ns, self.ncmax]))
        np.save(p / '_phy_spikes_subset.spikes.npy', np.zeros([self.ns]))
        np.save(p / '_phy_spikes_subset.waveforms.npy', np.zeros(
            [self.ns, self.nsamp, self.ncmax])
        )

        _write_tsv_simple(p / 'cluster_group.tsv', 'group', {2: 'good', 3: 'mua', 5: 'noise'})
        _write_tsv_simple(p / 'cluster_Amplitude.tsv', field_name='Amplitude',
                          data={str(n): np.random.rand() * 120 for n in np.arange(self.nt)})
        with open(p / 'probes.description.txt', 'w+') as fid:
            fid.writelines(['label\n'])

        # Raw data
        self.dat_path = p / 'rawdata.npy'
        np.save(self.dat_path, np.random.normal(size=(self.ncd, self.nc)))

        # LFP data.
        lfdata = (100 * np.random.normal(size=(1000, self.nc))).astype(np.int16)
        with (p / 'mydata.lf.bin').open('wb') as f:
            lfdata.tofile(f)

        self.files = os.listdir(self.tmp_dir)

    def _load(self, fn):
        p = Path(self.tmp_dir)
        return _load(p / fn)


@fixture
def dataset(tempdir):
    return Dataset(tempdir)


def test_ephys_1(dataset):
    assert dataset._load('spike_times.npy').shape == (dataset.ns,)
    assert dataset._load('spike_clusters.npy').shape == (dataset.ns,)
    assert dataset._load('amplitudes.npy').shape == (dataset.ns,)
    assert dataset._load('channel_positions.npy').shape == (dataset.nc, 2)
    assert dataset._load('templates.npy').shape == (dataset.nt, 50, dataset.nc)
    assert dataset._load('channel_map.npy').shape == (dataset.nc, 1)
    assert dataset._load('channel_probe.npy').shape == (dataset.nc,)
    assert len(dataset._load('cluster_group.tsv')) == 3
    assert dataset._load('rawdata.npy').shape == (1000, dataset.nc)
    assert dataset._load('mydata.lf.bin').shape == (1000 * dataset.nc,)
    assert dataset._load('whitening_mat.npy').shape == (dataset.nc, dataset.nc)
    assert dataset._load('_phy_spikes_subset.channels.npy').shape == (dataset.ns, dataset.ncmax)
    assert dataset._load('_phy_spikes_subset.spikes.npy').shape == (dataset.ns,)
    assert dataset._load('_phy_spikes_subset.waveforms.npy').shape == (
        (dataset.ns, dataset.nsamp, dataset.ncmax)
    )


def test_spike_depths(dataset):
    path = Path(dataset.tmp_dir)
    out_path = path / 'alf'

    mtemp = TemplateModel(
        dir_path=path, dat_path=dataset.dat_path, sample_rate=2000, n_channels_dat=dataset.nc)

    # create some sparse PC features
    n_subch = int(np.round(mtemp.n_channels / 2) - 1)
    pc_features = np.zeros((mtemp.n_spikes, n_subch, 3))
    close_channels = np.meshgrid(np.ones(mtemp.n_templates), np.arange(n_subch))[1]
    chind = mtemp.templates_channels
    chind = mtemp.templates_channels + close_channels * ((chind < 5) * 2 - 1)
    pc_features_ind = chind.transpose()
    # all PCs max between first and second channel
    pc_features[:, 0, 0] = 1
    pc_features[:, 1, 0] = 0.5
    print(mtemp.templates_channels)
    print(mtemp.templates_waveforms_durations)
    # add some depth information
    mtemp.channel_positions[:, 1] = mtemp.channel_positions[:, 0] + 10
    np.save(path / 'pc_features.npy', np.swapaxes(pc_features, 2, 1))
    np.save(path / 'pc_feature_ind.npy', pc_features_ind)
    np.save(path / 'channel_positions.npy', mtemp.channel_positions)

    model = TemplateModel(
        dir_path=path, dat_path=dataset.dat_path, sample_rate=2000, n_channels_dat=dataset.nc)

    c = EphysAlfCreator(model)
    shutil.rmtree(out_path, ignore_errors=True)
    c.convert(out_path)
    sd = np.load(next(out_path.glob('spikes.depths.npy')))
    sd_ = model.channel_positions[model.templates_channels[model.spike_templates], 1]
    assert np.all(np.abs(sd - sd_) <= 0.5)


def test_creator(dataset):
    _FILE_CREATES = (
        'spikes.times*.npy',
        'spikes.depths*.npy',
        'spikes.samples*.npy',
        'clusters.uuids*.csv',
        'clusters.amps*.npy',
        'clusters.channels*.npy',
        'clusters.depths*.npy',
        'clusters.peakToTrough*.npy',
        'clusters.waveforms*.npy',
        'clusters.waveformsChannels*.npy',
        'channels.localCoordinates*.npy',
        'channels.rawInd*.npy',
        'templates.waveforms*.npy',
        'templates.waveformsChannels*.npy',
    )
    path = Path(dataset.tmp_dir)
    out_path = path / 'alf'

    model = TemplateModel(
        dir_path=path, dat_path=dataset.dat_path, sample_rate=2000, n_channels_dat=dataset.nc)

    c = EphysAlfCreator(model)
    with raises(IOError):
        c.convert(dataset.tmp_dir)

    def check_conversion_output():
        # Check all renames.
        for old, new, _ in _FILE_RENAMES:
            if (path / old).exists():
                pattern = f'{Path(new).stem}*{Path(new).suffix}'
                assert next(out_path.glob(pattern)).exists()

        new_files = []
        for new in _FILE_CREATES:
            f = next(out_path.glob(new))
            new_files.append(f)
            assert f.exists()

        # makes sure the output dimensions match (especially clusters which should be 4)
        cl_shape = []
        for f in new_files:
            if f.name.startswith('clusters.') and f.name.endswith('.npy'):
                cl_shape.append(np.load(f).shape[0])
            elif f.name.startswith('clusters.') and f.name.endswith('.csv'):
                with open(f) as fid:
                    cl_shape.append(len(fid.readlines()) - 1)
        sp_shape = [np.load(f).shape[0] for f in new_files if f.name.startswith('spikes.')]
        ch_shape = [np.load(f).shape[0] for f in new_files if f.name.startswith('channels.')]

        assert len(set(cl_shape)) == 1
        assert len(set(sp_shape)) == 1
        assert len(set(ch_shape)) == 1

        dur = np.load(next(out_path.glob('clusters.peakToTrough*.npy')))
        assert np.all(dur == np.array([-9.5, 3., 13., -4.5, -2.5]))

    def read_after_write():
        model = TemplateModel(dir_path=out_path, dat_path=dataset.dat_path,
                              sample_rate=2000, n_channels_dat=dataset.nc)

        np.all(model.spike_templates == c.model.spike_templates)
        np.all(model.spike_times == c.model.spike_times)
        np.all(model.spike_samples == c.model.spike_samples)

    # test a straight export, make sure we can reload the data
    shutil.rmtree(out_path, ignore_errors=True)
    c.convert(out_path)
    check_conversion_output()
    read_after_write()

    # test with a label after the attribute name
    shutil.rmtree(out_path)
    c.convert(out_path, label='probe00')
    check_conversion_output()
    read_after_write()


def test_merger(dataset):

    path = Path(dataset.tmp_dir)
    out_path = path / 'alf'

    model = TemplateModel(
        dir_path=path, dat_path=dataset.dat_path, sample_rate=2000, n_channels_dat=dataset.nc)

    c = EphysAlfCreator(model)
    c.convert(out_path)

    model.close()

    # path.joinpath('_phy_spikes_subset.channels.npy').unlink()
    # path.joinpath('_phy_spikes_subset.waveforms.npy').unlink()
    # path.joinpath('_phy_spikes_subset.spikes.npy').unlink()

    out_path_merge = path / 'alf_merge'
    spike_clusters = dataset._load('spike_clusters.npy')
    clu, n_clu = np.unique(spike_clusters, return_counts=True)

    # merge the first two clusters
    merge_clu = clu[0:2]
    spike_clusters[np.bitwise_or(spike_clusters == clu[0],
                                 spike_clusters == clu[1])] = np.max(clu) + 1
    # split the cluster with the most spikes
    split_clu = clu[-1]
    idx = np.where(spike_clusters == split_clu)[0]
    spike_clusters[idx[0:int(n_clu[-1] / 2)]] = np.max(clu) + 2
    spike_clusters[idx[int(n_clu[-1] / 2):]] = np.max(clu) + 3

    np.save(path / 'spike_clusters.npy', spike_clusters)

    model = TemplateModel(
        dir_path=path, dat_path=dataset.dat_path, sample_rate=2000, n_channels_dat=dataset.nc)

    c = EphysAlfCreator(model)
    c.convert(out_path_merge)

    # Test that the split are the same for the expected datasets
    clu_old = np.load(next(out_path.glob('clusters.peakToTrough.npy')))
    clu_new = np.load(next(out_path_merge.glob('clusters.peakToTrough.npy')))
    assert clu_old[split_clu] == clu_new[np.max(clu) + 2]
    assert clu_old[split_clu] == clu_new[np.max(clu) + 3]

    assert np.isnan([clu_new[split_clu]])[0]
    assert np.isnan([clu_new[merge_clu[0]]])[0]
    assert np.isnan([clu_new[merge_clu[1]]])[0]

    clu_old = np.load(next(out_path.glob('clusters.channels.npy')))
    clu_new = np.load(next(out_path_merge.glob('clusters.channels.npy')))
    assert clu_old[split_clu] == clu_new[np.max(clu) + 2]
    assert clu_old[split_clu] == clu_new[np.max(clu) + 3]
    assert clu_new[split_clu] == 0
    assert clu_new[merge_clu[0]] == 0
    assert clu_new[merge_clu[1]] == 0

    clu_old = np.load(next(out_path.glob('clusters.depths.npy')))
    clu_new = np.load(next(out_path_merge.glob('clusters.depths.npy')))
    assert clu_old[split_clu] == clu_new[np.max(clu) + 2]
    assert clu_old[split_clu] == clu_new[np.max(clu) + 3]
    assert np.isnan([clu_new[split_clu]])[0]
    assert np.isnan([clu_new[merge_clu[0]]])[0]
    assert np.isnan([clu_new[merge_clu[1]]])[0]

    clu_old = np.load(next(out_path.glob('clusters.waveformsChannels.npy')))
    clu_new = np.load(next(out_path_merge.glob('clusters.waveformsChannels.npy')))
    assert np.array_equal(clu_old[split_clu], clu_new[np.max(clu) + 2])
    assert np.array_equal(clu_old[split_clu], clu_new[np.max(clu) + 3])



# -*- coding: utf-8 -*-

"""Tests of testing utility functions."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

from copy import deepcopy
import logging

import numpy as np

from ..testing import captured_output, captured_logging, _assert_equal

logger = logging.getLogger('phylib')


#------------------------------------------------------------------------------
# Tests
#------------------------------------------------------------------------------

def test_logging_1():
    print()
    logger.setLevel(5)
    logger.log(5, "level 5")
    logger.log(10, "debug")
    logger.log(20, "info")
    logger.log(30, "warning")
    logger.log(40, "error")


def test_captured_output():
    with captured_output() as (out, err):
        print('Hello world!')
    assert out.getvalue().strip() == 'Hello world!'


def test_captured_logging():
    handlers = logger.handlers
    with captured_logging() as buf:
        logger.debug('Hello world!')
    assert 'Hello world!' in buf.getvalue()
    assert logger.handlers == handlers


def test_assert_equal():
    d = {'a': {'b': np.random.rand(5), 3: 'c'}, 'b': 2.}
    d_bis = deepcopy(d)
    d_bis['a']['b'] = d_bis['a']['b'] + 1e-16
    _assert_equal(d, d_bis)



# -*- coding: utf-8 -*-

"""Plotting utilities."""


#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

from functools import partial
import logging

import numpy as np

logger = logging.getLogger(__name__)


#------------------------------------------------------------------------------
# Common probe layouts
#------------------------------------------------------------------------------

def linear_positions(n_channels):
    """Linear channel positions along the vertical axis."""
    return np.c_[np.zeros(n_channels),
                 np.linspace(0., 1., n_channels)]


def staggered_positions(n_channels):
    """Generate channel positions for a staggered probe."""
    i = np.arange(n_channels - 1)
    x, y = (-1) ** i * (5 + i), 10 * (i + 1)
    pos = np.flipud(np.r_[np.zeros((1, 2)), np.c_[x, y]])
    return pos


#------------------------------------------------------------------------------
# Box positioning
#------------------------------------------------------------------------------

def range_transform(from_bounds, to_bounds, positions, do_offset=True):
    """Transform for a rectangle to another."""
    from_bounds = np.asarray(from_bounds)
    to_bounds = np.asarray(to_bounds)
    positions = np.asarray(positions)

    assert from_bounds.ndim == to_bounds.ndim == positions.ndim == 2

    f0 = from_bounds[..., :2]
    f1 = from_bounds[..., 2:]
    t0 = to_bounds[..., :2]
    t1 = to_bounds[..., 2:]

    # Degenerate axes are extended maximally.
    for z0, z1 in ((f0, f1), (t0, t1)):
        for i in range(2):
            ind = np.abs(z0[:, i] - z1[:, i]) < 1e-8
            z0[ind, i] = -1
            z1[ind, i] = +1

    d = (f1 - f0)
    d[d == 0] = 1

    out = positions.copy()
    if do_offset:
        out -= f0.astype(out.dtype)
    out *= ((t1 - t0) / d).astype(out.dtype)
    if do_offset:
        out += t0.astype(out.dtype)
    return out


def _boxes_overlap(x0, y0, x1, y1):
    """Return whether a set of boxes, defined by their 2D corners, overlap or  not."""
    assert x0.ndim == y0.ndim == y0.ndim == y1.ndim == 2
    n = len(x0)
    overlap_matrix = ((x0 < x1.T) & (x1 > x0.T) & (y0 < y1.T) & (y1 > y0.T))
    overlap_matrix[np.arange(n), np.arange(n)] = False
    return np.any(overlap_matrix.ravel())


def _binary_search(f, xmin, xmax, eps=1e-9):
    """Return the largest x such f(x) is True."""
    middle = (xmax + xmin) / 2.
    while xmax - xmin > eps:
        assert xmin < xmax
        middle = (xmax + xmin) / 2.
        if f(xmax):
            return xmax
        if not f(xmin):
            return xmin
        if f(middle):
            xmin = middle
        else:
            xmax = middle
    return middle


def _find_box_size(x, y, ar=.5, margin=0):
    """Return the maximum (half) box size such that boxes centered around box positions
    do not overlap."""
    if x.ndim == 1:
        x = x[:, np.newaxis]
    if y.ndim == 1:
        y = y[:, np.newaxis]
    logger.log(5, "Get box size for %d points.", len(x))
    # Deal with degenerate x case.
    xmin, xmax = x.min(), x.max()
    if xmin == xmax:
        # If all positions are vertical, the width can be maximum.
        wmax = 1.
    else:
        wmax = xmax - xmin

    def f1(w, keep_aspect_ratio=True, h=None):
        """Return true if the configuration with the current box size
        is non-overlapping."""
        # NOTE: w|h are the *half* width|height.
        if keep_aspect_ratio:
            h = w * ar  # fixed aspect ratio
        return not _boxes_overlap(x - w, y - h, x + w, y + h)

    # Find the largest box size leading to non-overlapping boxes.
    w = _binary_search(f1, 0, wmax)
    w = w * (1 - margin)  # margin
    # Clip the half-width.
    h = w * ar  # aspect ratio

    # Extend the boxes horizontally as much as possible.
    w = _binary_search(partial(f1, keep_aspect_ratio=False, h=h), w, wmax)
    w = w * (1 - margin)  # margin

    return w, h


def get_non_overlapping_boxes(box_pos):
    """Normalize box positions and return a convenient half box size."""
    box_pos = np.asarray(box_pos)
    assert box_pos.ndim == 2
    assert box_pos.shape[1] == 2
    # Renormalize box_pos.
    mx, my = box_pos.min(axis=0)
    Mx, My = box_pos.max(axis=0)
    box_pos = range_transform([[mx, my, Mx, My]], [[-1, -1, +1, +1]], box_pos)
    # Compute box size.
    x, y = box_pos.T
    w, h = _find_box_size(x, y, margin=.1)
    # Renormalize again so that the boxes fit inside the view.
    mx, my = np.min(box_pos - np.array([[w, h]]), axis=0)
    Mx, My = np.max(box_pos + np.array([[w, h]]), axis=0)
    b1 = [[mx, my, Mx, My]]
    b2 = [[-1, -1, 1, 1]]
    box_pos = range_transform(b1, b2, box_pos)
    w, h = range_transform(b1, b2, [[w, h]], do_offset=False).ravel()
    w *= .95
    h *= .9
    logger.log(5, "Found box size %s.", (w, h))
    return box_pos, (w, h)


def get_closest_box(pos, box_pos, box_size):
    """Return the box closest to a given point."""
    # box_size is the half size
    # see https://gamedev.stackexchange.com/a/44496
    w, h = box_size
    x, y = pos
    px, py = box_pos.T
    dx = np.maximum(np.abs(px - x) - w, 0)
    dy = np.maximum(np.abs(py - y) - h, 0)
    d = dx * dx + dy * dy
    return np.argmin(d)


#------------------------------------------------------------------------------
# Data bounds utilities
#------------------------------------------------------------------------------

def _get_data_bounds(data_bounds, pos=None, length=None):
    """"Prepare data bounds, possibly using min/max of the data."""
    if data_bounds is None or (isinstance(data_bounds, str) and data_bounds == 'auto'):
        if pos is not None and len(pos):
            m, M = pos.min(axis=0), pos.max(axis=0)
            data_bounds = [m[0], m[1], M[0], M[1]]
        else:
            data_bounds = [-1, -1, 1, 1]
    data_bounds = np.atleast_2d(data_bounds)

    ind_x = data_bounds[:, 0] == data_bounds[:, 2]
    ind_y = data_bounds[:, 1] == data_bounds[:, 3]
    if np.sum(ind_x):
        data_bounds[ind_x, 0] -= 1
        data_bounds[ind_x, 2] += 1
    if np.sum(ind_y):
        data_bounds[ind_y, 1] -= 1
        data_bounds[ind_y, 3] += 1

    # Extend the data_bounds if needed.
    if length is None:
        length = pos.shape[0] if pos is not None else 1
    if data_bounds.shape[0] == 1:
        data_bounds = np.tile(data_bounds, (length, 1))

    # Check the shape of data_bounds.
    assert data_bounds.shape == (length, 4)

    assert data_bounds.ndim == 2
    assert data_bounds.shape[1] == 4
    assert np.all(data_bounds[:, 0] < data_bounds[:, 2])
    assert np.all(data_bounds[:, 1] < data_bounds[:, 3])

    return data_bounds



# -*- coding: utf-8 -*-

"""py.test utilities."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import logging
import os
from pathlib import Path
import tempfile
import shutil
import warnings

import numpy as np
from pytest import fixture

from phylib import add_default_handler


#------------------------------------------------------------------------------
# Common fixtures
#------------------------------------------------------------------------------

logger = logging.getLogger('phylib')
logger.setLevel(10)
add_default_handler(5, logger=logger)

# Fix the random seed in the tests.
np.random.seed(2015)

warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")


@fixture
def tempdir():
    curdir = os.getcwd()
    tempdir = tempfile.mkdtemp()
    os.chdir(tempdir)
    yield Path(tempdir)
    os.chdir(curdir)
    shutil.rmtree(tempdir)



# -*- coding: utf-8 -*-
# flake8: noqa

"""Utilities for large-scale ephys data analysis."""


#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import atexit
import logging
import os.path as op

from .utils._misc import _git_version
from .utils.event import connect, unconnect, emit


#------------------------------------------------------------------------------
# Global variables and functions
#------------------------------------------------------------------------------

__author__ = 'Cyrille Rossant'
__email__ = 'cyrille.rossant at gmail.com'
__version__ = '2.6.2'
__version_git__ = __version__ + _git_version()


# Set a null handler on the root logger
logger = logging.getLogger('phylib')
logger.setLevel(logging.DEBUG)
logger.addHandler(logging.NullHandler())


_logger_fmt = '%(asctime)s.%(msecs)03d [%(levelname)s] %(caller)s %(message)s'
_logger_date_fmt = '%H:%M:%S'


class _Formatter(logging.Formatter):
    color_codes = {'L': '94', 'D': '90', 'I': '0', 'W': '33', 'E': '31'}

    def format(self, record):
        # Only keep the first character in the level name.
        record.levelname = record.levelname[0]
        filename = op.splitext(op.basename(record.pathname))[0]
        record.caller = '{:s}:{:d}'.format(filename, record.lineno).ljust(20)
        message = super(_Formatter, self).format(record)
        color_code = self.color_codes.get(record.levelname, '90')
        message = '\33[%sm%s\33[0m' % (color_code, message)
        return message


def add_default_handler(level='INFO', logger=logger):
    handler = logging.StreamHandler()
    handler.setLevel(level)

    formatter = _Formatter(fmt=_logger_fmt, datefmt=_logger_date_fmt)
    handler.setFormatter(formatter)

    logger.addHandler(handler)


def _add_log_file(filename):  # pragma: no cover
    """Create a log file with DEBUG level."""
    handler = logging.FileHandler(str(filename))
    handler.setLevel(logging.DEBUG)
    formatter = _Formatter(fmt=_logger_fmt, datefmt=_logger_date_fmt)
    handler.setFormatter(formatter)
    logging.getLogger('phy').addHandler(handler)


@atexit.register
def on_exit():  # pragma: no cover
    # Close the logging handlers.
    for handler in logger.handlers:
        handler.close()
        logger.removeHandler(handler)


def test():  # pragma: no cover
    """Run the full testing suite of phylib."""
    import pytest
    pytest.main()



# -*- coding: utf-8 -*-

"""Testing the BaseEphysTraces class."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import logging

import numpy as np
from numpy.testing import assert_equal as ae
from numpy.testing import assert_allclose as ac
import mtscomp
from pytest import raises, fixture, mark

from phylib.utils import Bunch
from ..traces import (
    _get_subitems, _get_chunk_bounds,
    get_ephys_reader, BaseEphysReader, extract_waveforms, export_waveforms, RandomEphysReader,
    get_spike_waveforms)

logger = logging.getLogger(__name__)


#------------------------------------------------------------------------------
# Test utils
#------------------------------------------------------------------------------

def test_get_subitems():
    bounds = [0, 2, 5]

    def _a(x, y):
        res = _get_subitems(bounds, x)
        res = [(chk, val.tolist() if isinstance(val, np.ndarray) else val) for chk, val in res]
        assert res == y

    _a(-1, [(1, 2)])
    _a(0, [(0, 0)])
    _a(2, [(1, 0)])
    _a(4, [(1, 2)])
    with raises(IndexError):
        _a(5, [])

    _a(slice(None, None, None), [(0, slice(0, 2, 1)), (1, slice(0, 3, 1))])

    _a(slice(1, None, 1), [(0, slice(1, 2, 1)), (1, slice(0, 3, 1))])

    _a(slice(2, None, 1), [(1, slice(0, 3, 1))])
    _a(slice(3, None, 1), [(1, slice(1, 3, 1))])
    _a(slice(5, None, 1), [])

    _a(slice(0, 4, 1), [(0, slice(0, 2, 1)), (1, slice(0, 2, 1))])
    _a(slice(1, 2, 1), [(0, slice(1, 2, 1))])
    _a(slice(1, -1, 1), [(0, slice(1, 2, 1)), (1, slice(0, 2, 1))])
    _a(slice(-2, -1, 1), [(1, slice(1, 2, 1))])

    _a([0], [(0, [0])])
    _a([2], [(1, [0])])
    _a([4], [(1, [2])])
    with raises(IndexError):
        _a([5], [])

    _a([0, 1], [(0, [0, 1])])
    _a([0, 2], [(0, [0]), (1, [0])])
    _a([0, 3], [(0, [0]), (1, [1])])
    with raises(IndexError):
        _a([0, 5], [(0, [0])])
    _a([3, 4], [(1, [1, 2])])

    _a(([3, 4], None), [(1, [1, 2])])


def test_get_chunk_bounds():
    def _a(x, y, z):
        assert _get_chunk_bounds(x, y) == z

    _a([3], 2, [0, 2, 3])
    _a([3], 3, [0, 3])
    _a([3], 4, [0, 3])

    _a([3, 2], 2, [0, 2, 3, 5])
    _a([3, 2], 3, [0, 3, 5])

    _a([3, 7, 5], 4, [0, 3, 7, 10, 14, 15])
    _a([3, 7, 6], 4, [0, 3, 7, 10, 14, 16])

    _a([3, 7, 5], 10, [0, 3, 10, 15])


#------------------------------------------------------------------------------
# Test ephys reader
#------------------------------------------------------------------------------

@fixture
def arr():
    return np.random.randn(2000, 10)


@fixture(params=[10000, 1000, 100])
def sample_rate(request):
    return request.param


@fixture(params=['numpy', 'npy', 'flat', 'flat_concat', 'mtscomp', 'mtscomp_reader'])
def traces(request, tempdir, arr, sample_rate):
    if request.param == 'numpy':
        return get_ephys_reader(arr, sample_rate=sample_rate)

    elif request.param == 'npy':
        path = tempdir / 'data.npy'
        np.save(path, arr)
        return get_ephys_reader(path, sample_rate=sample_rate)

    elif request.param == 'flat':
        path = tempdir / 'data.bin'
        with open(path, 'wb') as f:
            arr.tofile(f)
        return get_ephys_reader(
            path, sample_rate=sample_rate, dtype=arr.dtype, n_channels=arr.shape[1])

    elif request.param == 'flat_concat':
        path0 = tempdir / 'data0.bin'
        with open(path0, 'wb') as f:
            arr[:arr.shape[0] // 2, :].tofile(f)
        path1 = tempdir / 'data1.bin'
        with open(path1, 'wb') as f:
            arr[arr.shape[0] // 2:, :].tofile(f)
        return get_ephys_reader(
            [path0, path1], sample_rate=sample_rate, dtype=arr.dtype, n_channels=arr.shape[1])

    elif request.param in ('mtscomp', 'mtscomp_reader'):
        path = tempdir / 'data.bin'
        with open(path, 'wb') as f:
            arr.tofile(f)
        out = tempdir / 'data.cbin'
        outmeta = tempdir / 'data.ch'
        mtscomp.compress(
            path, out, outmeta, sample_rate=sample_rate,
            n_channels=arr.shape[1], dtype=arr.dtype,
            n_threads=1, check_after_compress=False, quiet=True)
        reader = mtscomp.decompress(out, outmeta, check_after_decompress=False, quiet=True)
        if request.param == 'mtscomp':
            return get_ephys_reader(reader)
        else:
            return get_ephys_reader(out)


def test_ephys_reader_1(tempdir, arr, traces, sample_rate):
    assert isinstance(traces, BaseEphysReader)
    assert traces.dtype == arr.dtype
    assert traces.ndim == 2
    assert traces.shape == arr.shape
    assert traces.n_samples == arr.shape[0]
    assert traces.n_channels == arr.shape[1]
    assert traces.n_parts in (1, 2)
    assert traces.duration == arr.shape[0] / sample_rate
    assert len(traces.part_bounds) == traces.n_parts + 1
    assert len(traces.chunk_bounds) == traces.n_chunks + 1

    ac(traces[:], arr)

    def _a(f):
        ac(f(traces)[:], f(arr))

    _a(lambda x: x[:, ::-1])

    _a(lambda x: x + 1)
    _a(lambda x: 1 + x)

    _a(lambda x: x - 1)
    _a(lambda x: 1 - x)

    _a(lambda x: x * 2)
    _a(lambda x: 2 * x)

    _a(lambda x: x ** 2)
    _a(lambda x: 2 ** x)

    _a(lambda x: x / 2)
    _a(lambda x: 2 / x)

    _a(lambda x: x / 2.)
    _a(lambda x: 2. / x)

    _a(lambda x: x // 2)
    _a(lambda x: 2 // x)

    _a(lambda x: +x)
    _a(lambda x: -x)

    _a(lambda x: -x[:, [1, 3, 5]])

    _a(lambda x: 1 + x * 2)
    _a(lambda x: 1 + (2 * x))
    _a(lambda x: -x * 2)

    _a(lambda x: x[::1])
    _a(lambda x: x[::1, :])
    _a(lambda x: x[::1, 1:5])
    _a(lambda x: x[::1, ::3])


def test_ephys_random(sample_rate):
    reader = RandomEphysReader(2000, 10, sample_rate=sample_rate)
    assert reader[:10].shape == (10, 10)
    assert reader[:].shape == (2000, 10)
    assert reader[0].shape == (1, 10)
    assert reader[10:20].shape == (10, 10)
    assert reader[[1, 3, 5]].shape == (3, 10)
    assert reader[[1, 3, 5], :].shape == (3, 10)
    assert reader[[1, 3, 5], ::2].shape == (3, 5)
    assert reader[[1, 3, 5], [0, 2, 4]].shape == (3, 3)
    assert reader[0:-1].shape == (1999, 10)
    assert reader[-10:-1].shape == (9, 10)


def test_get_spike_waveforms():
    ns, nsw, nc = 8, 5, 3

    w = np.random.rand(ns, nsw, nc)
    s = np.arange(1, 1 + 2 * ns, 2)
    c = np.tile(np.array([1, 2, 3]), (ns, 1))

    assert w.shape == (ns, nsw, nc)
    assert s.shape == (ns,)
    assert c.shape == (ns, nc)

    sw = Bunch(waveforms=w, spike_ids=s, spike_channels=c)
    out = get_spike_waveforms([5, 1, 3], [2, 1], spike_waveforms=sw, n_samples_waveforms=nsw)

    expected = w[[2, 0, 1], ...][..., [1, 0]]
    ae(out, expected)


@mark.parametrize('do_export', [False, True])
@mark.parametrize('do_cache', [False, True])
def test_waveform_extractor(tempdir, arr, traces, sample_rate, do_export, do_cache):
    data = arr

    nsw = 20
    channel_ids = [1, 3, 5]
    spike_samples = [5, 25, 100, 1000, 1995]
    spike_ids = np.arange(len(spike_samples))
    spike_channels = np.array([channel_ids] * len(spike_samples))

    # Export waveforms into a npy file.
    if do_export:
        export_waveforms(
            tempdir / 'waveforms.npy', traces, spike_samples, spike_channels,
            n_samples_waveforms=nsw, cache=do_cache)
        w = np.load(tempdir / 'waveforms.npy')
    # Extract waveforms directly.
    else:
        w = extract_waveforms(traces, spike_samples, channel_ids, n_samples_waveforms=nsw)

    assert w.dtype == data.dtype == traces.dtype

    spike_waveforms = Bunch(
        spike_ids=spike_ids,
        spike_channels=spike_channels,
        waveforms=w,
    )

    ww = get_spike_waveforms(
        spike_ids, channel_ids, spike_waveforms=spike_waveforms,
        n_samples_waveforms=nsw)
    ae(w, ww)

    assert np.all(w[0, :5, :] == 0)
    ac(w[0, 5:, :], data[0:15, [1, 3, 5]])

    ac(w[1, ...], data[15:35, [1, 3, 5]])
    ac(w[2, ...], data[90:110, [1, 3, 5]])
    ac(w[3, ...], data[990:1010, [1, 3, 5]])

    assert np.all(w[4, -5:, :] == 0)
    ac(w[4, :-5, :], data[-15:, [1, 3, 5]])



# -*- coding: utf-8 -*-

"""Tests of array utility functions."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

from pathlib import Path

import numpy as np
from pytest import raises

from ..array import (
    _unique, _normalize, _index_of, _spikes_in_clusters, _spikes_per_cluster,
    _flatten_per_cluster, get_closest_clusters, _get_data_lim, _flatten, _clip,
    chunk_bounds, excerpts, data_chunk, grouped_mean, SpikeSelector,
    get_excerpts, _range_from_slice, _pad, _get_padded,
    read_array, write_array)
from phylib.utils._types import _as_array
from phylib.utils.testing import _assert_equal as ae
from ..mock import artificial_spike_clusters, artificial_spike_samples


#------------------------------------------------------------------------------
# Test utility functions
#------------------------------------------------------------------------------

def test_clip():
    assert _clip(-1, 0, 1) == 0


def test_range_from_slice():
    """Test '_range_from_slice'."""

    class _SliceTest(object):
        """Utility class to make it more convenient to test slice objects."""
        def __init__(self, **kwargs):
            self._kwargs = kwargs

        def __getitem__(self, item):
            if isinstance(item, slice):
                return _range_from_slice(item, **self._kwargs)

    with raises(ValueError):
        _SliceTest()[:]
    with raises(ValueError):
        _SliceTest()[1:]
    ae(_SliceTest()[:5], [0, 1, 2, 3, 4])
    ae(_SliceTest()[1:5], [1, 2, 3, 4])

    with raises(ValueError):
        _SliceTest()[::2]
    with raises(ValueError):
        _SliceTest()[1::2]
    ae(_SliceTest()[1:5:2], [1, 3])

    with raises(ValueError):
        _SliceTest(start=0)[:]
    with raises(ValueError):
        _SliceTest(start=1)[:]
    with raises(ValueError):
        _SliceTest(step=2)[:]

    ae(_SliceTest(stop=5)[:], [0, 1, 2, 3, 4])
    ae(_SliceTest(start=1, stop=5)[:], [1, 2, 3, 4])
    ae(_SliceTest(stop=5)[1:], [1, 2, 3, 4])
    ae(_SliceTest(start=1)[:5], [1, 2, 3, 4])
    ae(_SliceTest(start=1, step=2)[:5], [1, 3])
    ae(_SliceTest(start=1)[:5:2], [1, 3])

    ae(_SliceTest(length=5)[:], [0, 1, 2, 3, 4])
    with raises(ValueError):
        _SliceTest(length=5)[:3]
    ae(_SliceTest(length=5)[:10], [0, 1, 2, 3, 4])
    ae(_SliceTest(length=5)[:5], [0, 1, 2, 3, 4])
    ae(_SliceTest(start=1, length=5)[:], [1, 2, 3, 4, 5])
    ae(_SliceTest(start=1, length=5)[:6], [1, 2, 3, 4, 5])
    with raises(ValueError):
        _SliceTest(start=1, length=5)[:4]
    ae(_SliceTest(start=1, step=2, stop=5)[:], [1, 3])
    ae(_SliceTest(start=1, stop=5)[::2], [1, 3])
    ae(_SliceTest(stop=5)[1::2], [1, 3])


def test_pad():
    arr = np.random.rand(10, 3)

    ae(_pad(arr, 0, 'right'), arr[:0, :])
    ae(_pad(arr, 3, 'right'), arr[:3, :])
    ae(_pad(arr, 9), arr[:9, :])
    ae(_pad(arr, 10), arr)

    ae(_pad(arr, 12, 'right')[:10, :], arr)
    ae(_pad(arr, 12)[10:, :], np.zeros((2, 3)))

    ae(_pad(arr, 0, 'left'), arr[:0, :])
    ae(_pad(arr, 3, 'left'), arr[7:, :])
    ae(_pad(arr, 9, 'left'), arr[1:, :])
    ae(_pad(arr, 10, 'left'), arr)

    ae(_pad(arr, 12, 'left')[2:, :], arr)
    ae(_pad(arr, 12, 'left')[:2, :], np.zeros((2, 3)))

    with raises(ValueError):
        _pad(arr, -1)


def test_get_padded():
    arr = np.array([1, 2, 3])[:, np.newaxis]

    with raises(RuntimeError):
        ae(_get_padded(arr, -2, 5).ravel(), [1, 2, 3, 0, 0])
    ae(_get_padded(arr, 1, 2).ravel(), [2])
    ae(_get_padded(arr, 0, 5).ravel(), [1, 2, 3, 0, 0])
    ae(_get_padded(arr, -2, 3).ravel(), [0, 0, 1, 2, 3])


def test_get_data_lim():
    arr = np.random.rand(10, 5)
    assert 0 < _get_data_lim(arr) < 1
    assert 0 < _get_data_lim(arr, 2) < 1


def test_unique():
    """Test _unique() function"""
    _unique([])

    n_spikes = 300
    n_clusters = 3
    spike_clusters = artificial_spike_clusters(n_spikes, n_clusters)
    ae(_unique(spike_clusters), np.arange(n_clusters))


def test_normalize():
    """Test _normalize() function."""

    n_channels = 10
    positions = 1 + 2 * np.random.randn(n_channels, 2)

    # Keep ration is False.
    positions_n = _normalize(positions)

    x_min, y_min = positions_n.min(axis=0)
    x_max, y_max = positions_n.max(axis=0)

    np.allclose(x_min, 0.)
    np.allclose(x_max, 1.)
    np.allclose(y_min, 0.)
    np.allclose(y_max, 1.)

    # Keep ratio is True.
    positions_n = _normalize(positions, keep_ratio=True)

    x_min, y_min = positions_n.min(axis=0)
    x_max, y_max = positions_n.max(axis=0)

    np.allclose(min(x_min, y_min), 0.)
    np.allclose(max(x_max, y_max), 1.)
    np.allclose(x_min + x_max, 1)
    np.allclose(y_min + y_max, 1)


def test_index_of():
    """Test _index_of."""
    arr = [36, 42, 42, 36, 36, 2, 42]
    lookup = _unique(arr)
    ae(_index_of(arr, lookup), [1, 2, 2, 1, 1, 0, 2])


def test_as_array():
    ae(_as_array(3), [3])
    ae(_as_array([3]), [3])
    ae(_as_array(3.), [3.])
    ae(_as_array([3.]), [3.])

    with raises(ValueError):
        _as_array(map)


def test_flatten():
    assert _flatten([[0, 1], [2]]) == [0, 1, 2]


def test_get_closest_clusters():
    out = get_closest_clusters(1, [0, 1, 2], lambda c, d: (d - c))
    assert [_ for _, __ in out] == [2, 1, 0]


#------------------------------------------------------------------------------
# Test read/save
#------------------------------------------------------------------------------

def test_read_write(tempdir):
    arr = np.arange(10).astype(np.float32)

    path = Path(tempdir) / 'test.npy'

    write_array(path, arr)
    ae(read_array(path), arr)
    ae(read_array(path, mmap_mode='r'), arr)


#------------------------------------------------------------------------------
# Test chunking
#------------------------------------------------------------------------------

def test_chunk_bounds():
    chunks = chunk_bounds(200, 100, overlap=20)

    assert next(chunks) == (0, 100, 0, 90)
    assert next(chunks) == (80, 180, 90, 170)
    assert next(chunks) == (160, 200, 170, 200)


def test_chunk():
    data = np.random.randn(200, 4)
    chunks = chunk_bounds(data.shape[0], 100, overlap=20)

    with raises(ValueError):
        data_chunk(data, (0, 0, 0))

    assert data_chunk(data, (0, 0)).shape == (0, 4)

    # Chunk 1.
    ch = next(chunks)
    d = data_chunk(data, ch)
    d_o = data_chunk(data, ch, with_overlap=True)

    ae(d_o, data[0:100])
    ae(d, data[0:90])

    # Chunk 2.
    ch = next(chunks)
    d = data_chunk(data, ch)
    d_o = data_chunk(data, ch, with_overlap=True)

    ae(d_o, data[80:180])
    ae(d, data[90:170])


def test_excerpts_1():
    bounds = [(start, end) for (start, end) in excerpts(100,
                                                        n_excerpts=3,
                                                        excerpt_size=10)]
    assert bounds == [(0, 10), (45, 55), (90, 100)]


def test_excerpts_2():
    bounds = [(start, end) for (start, end) in excerpts(10,
                                                        n_excerpts=3,
                                                        excerpt_size=10)]
    assert bounds == [(0, 10)]


def test_get_excerpts():
    data = np.random.rand(100, 2)
    subdata = get_excerpts(data, n_excerpts=10, excerpt_size=5)
    assert subdata.shape == (50, 2)
    ae(subdata[:5, :], data[:5, :])
    ae(subdata[-5:, :], data[-10:-5, :])

    data = np.random.rand(10, 2)
    subdata = get_excerpts(data, n_excerpts=10, excerpt_size=5)
    ae(subdata, data)

    data = np.random.rand(10, 2)
    subdata = get_excerpts(data, n_excerpts=1, excerpt_size=10)
    ae(subdata, data)

    assert len(get_excerpts(data, n_excerpts=0, excerpt_size=10)) == 0


#------------------------------------------------------------------------------
# Test spike clusters functions
#------------------------------------------------------------------------------

def test_spikes_in_clusters():
    """Test _spikes_in_clusters()."""

    n_spikes = 100
    n_clusters = 5
    spike_clusters = artificial_spike_clusters(n_spikes, n_clusters)

    ae(_spikes_in_clusters(spike_clusters, []), [])

    for i in range(n_clusters):
        assert np.all(spike_clusters[_spikes_in_clusters(spike_clusters, [i])] == i)

    clusters = [1, 2, 3]
    assert np.all(np.isin(
        spike_clusters[_spikes_in_clusters(spike_clusters, clusters)], clusters))


def test_spikes_per_cluster():
    """Test _spikes_per_cluster()."""

    n_spikes = 100
    n_clusters = 3
    spike_clusters = artificial_spike_clusters(n_spikes, n_clusters)

    assert not _spikes_per_cluster([])

    spikes_per_cluster = _spikes_per_cluster(spike_clusters)
    assert list(spikes_per_cluster.keys()) == list(range(n_clusters))

    for i in range(n_clusters):
        ae(spikes_per_cluster[i], np.sort(spikes_per_cluster[i]))
        assert np.all(spike_clusters[spikes_per_cluster[i]] == i)


def test_flatten_per_cluster():
    spc = {2: [2, 7, 11], 3: [3, 5], 5: []}
    arr = _flatten_per_cluster(spc)
    ae(arr, [2, 3, 5, 7, 11])


def test_grouped_mean():
    spike_clusters = np.array([2, 3, 2, 2, 5])
    arr = [9, -3, 10, 11, -5]
    ae(grouped_mean(arr, spike_clusters), [10, -3, -5])


#------------------------------------------------------------------------------
# Test spike selection
#------------------------------------------------------------------------------

def test_select_spikes_1():
    spike_times = np.array([0., 1., 2., 3.3, 4.4])
    spike_clusters = np.array([1, 2, 1, 2, 4])
    chunk_bounds = [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
    n_chunks_kept = 2
    cluster_ids = [1, 2, 4]
    spikes_ids_kept = [0, 1, 3]

    spc = _spikes_per_cluster(spike_clusters)
    ss = SpikeSelector(
        get_spikes_per_cluster=lambda cl: spc.get(cl, np.array([], dtype=np.int64)),
        spike_times=spike_times, chunk_bounds=chunk_bounds, n_chunks_kept=n_chunks_kept)
    ae(ss.chunks_kept, [0.0, 1.1, 3.3, 4.4])

    ae(ss(3, [], subset_chunks=True), [])
    ae(ss(3, [0], subset_chunks=True), [])
    ae(ss(3, [1], subset_chunks=True), [0])

    ae(ss(None, cluster_ids, subset_chunks=True), spikes_ids_kept)
    ae(ss(0, cluster_ids, subset_chunks=True), spikes_ids_kept)
    ae(ss(3, cluster_ids, subset_chunks=True), spikes_ids_kept)
    ae(ss(2, cluster_ids, subset_chunks=True), spikes_ids_kept)
    assert list(ss(1, cluster_ids, subset_chunks=True)) in [[0, 1], [0, 3]]

    ae(ss(2, cluster_ids, subset_spikes=[0, 1], subset_chunks=True), [0, 1])
    ae(ss(2, cluster_ids, subset_chunks=False), np.arange(5))


def test_select_spikes_2():
    n_spikes = 1000
    n_clusters = 10
    spike_times = artificial_spike_samples(n_spikes)
    spike_times = 10. * spike_times / spike_times.max()
    chunk_bounds = np.linspace(0.0, 10.0, 11)
    n_chunks_kept = 3
    chunks_kept = [0., 1., 4., 5., 8., 9.]
    spike_clusters = artificial_spike_clusters(n_spikes, n_clusters)

    spc = _spikes_per_cluster(spike_clusters)
    ss = SpikeSelector(
        get_spikes_per_cluster=lambda cl: spc.get(cl, np.array([], dtype=np.int64)),
        spike_times=spike_times, chunk_bounds=chunk_bounds, n_chunks_kept=n_chunks_kept)
    ae(ss.chunks_kept, chunks_kept)

    def _check_chunks(sid):
        chunk_ids = np.searchsorted(chunk_bounds, spike_times[sid], 'right') - 1
        ae(np.unique(chunk_ids), [0, 4, 8])

    # Select all spikes belonging to the kept chunks.
    sid = ss(n_spikes, np.arange(n_clusters), subset_chunks=True)
    _check_chunks(sid)

    # Select 10 spikes from each cluster.
    sid = ss(10, np.arange(n_clusters), subset_chunks=True)
    assert np.all(np.diff(sid) > 0)
    _check_chunks(sid)
    ae(np.bincount(spike_clusters[sid]), [10] * 10)



# -*- coding: utf-8 -*-

"""Cluster statistics."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import numpy as np


#------------------------------------------------------------------------------
# Cluster statistics
#------------------------------------------------------------------------------

def mean(x):
    """Return the mean of an array across the first dimension."""
    return x.mean(axis=0)


def get_unmasked_channels(mean_masks, min_mask=.25):
    """Return the unmasked channels (mean masks above a given threshold)."""
    return np.nonzero(mean_masks > min_mask)[0]


def get_mean_probe_position(mean_masks, site_positions):
    """Return the mean position of clusters on the probe, depending on the masks."""
    m = max(1, np.sum(mean_masks))
    return np.sum(site_positions * mean_masks[:, np.newaxis], axis=0) / m


def get_sorted_main_channels(mean_masks, unmasked_channels):
    """Weighted mean of the channels, weighted by the mean masks."""
    main_channels = np.argsort(mean_masks)[::-1]
    main_channels = np.array([c for c in main_channels
                              if c in unmasked_channels])
    return main_channels


#------------------------------------------------------------------------------
# Wizard measures
#------------------------------------------------------------------------------

def get_waveform_amplitude(mean_masks, mean_waveforms):
    """Return the amplitude of the waveforms on all channels."""

    assert mean_waveforms.ndim == 2
    n_samples, n_channels = mean_waveforms.shape

    assert mean_masks.ndim == 1
    assert mean_masks.shape == (n_channels,)

    mean_waveforms = mean_waveforms * mean_masks
    assert mean_waveforms.shape == (n_samples, n_channels)

    # Amplitudes.
    m, M = mean_waveforms.min(axis=0), mean_waveforms.max(axis=0)
    return M - m


def get_mean_masked_features_distance(
        mean_features_0, mean_features_1, mean_masks_0, mean_masks_1, n_features_per_channel=None):
    """Compute the distance between the mean masked features."""

    assert n_features_per_channel > 0

    mu_0 = mean_features_0.ravel()
    mu_1 = mean_features_1.ravel()

    omeg_0 = mean_masks_0
    omeg_1 = mean_masks_1

    omeg_0 = np.repeat(omeg_0, n_features_per_channel)
    omeg_1 = np.repeat(omeg_1, n_features_per_channel)

    d_0 = mu_0 * omeg_0
    d_1 = mu_1 * omeg_1

    return np.linalg.norm(d_0 - d_1)



# -*- coding: utf-8 -*-

"""Multi-electrode arrays."""


#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import itertools
from pathlib import Path

import numpy as np

from phylib.utils._types import _as_array
from phylib.utils._misc import read_python


#------------------------------------------------------------------------------
# PRB file utilities
#------------------------------------------------------------------------------

def _edges_to_adjacency_list(edges):
    """Convert a list of edges into an adjacency list."""
    adj = {}
    for i, j in edges:
        if i in adj:  # pragma: no cover
            ni = adj[i]
        else:
            ni = adj[i] = set()
        if j in adj:
            nj = adj[j]
        else:
            nj = adj[j] = set()
        ni.add(j)
        nj.add(i)
    return adj


def _adjacency_subset(adjacency, subset):
    return {c: [v for v in vals if v in subset]
            for (c, vals) in adjacency.items() if c in subset}


def _remap_adjacency(adjacency, mapping):
    remapped = {}
    for key, vals in adjacency.items():
        remapped[mapping[key]] = [mapping[i] for i in vals]
    return remapped


def _probe_positions(probe, group):
    """Return the positions of a probe channel group."""
    positions = probe['channel_groups'][group]['geometry']
    channels = _probe_channels(probe, group)
    return np.array([positions[channel] for channel in channels])


def _probe_channels(probe, group):
    """Return the list of channels in a channel group.

    The order is kept.

    """
    return probe['channel_groups'][group]['channels']


def _probe_adjacency_list(probe):
    """Return an adjacency list of a whole probe."""
    cgs = probe['channel_groups'].values()
    graphs = [cg['graph'] for cg in cgs]
    edges = list(itertools.chain(*graphs))
    adjacency_list = _edges_to_adjacency_list(edges)
    return adjacency_list


def _channels_per_group(probe):
    groups = probe['channel_groups'].keys()
    return {group: probe['channel_groups'][group]['channels']
            for group in groups}


def load_probe(name_or_path):
    """Load one of the built-in probes."""
    path = Path(name_or_path)
    # The argument can be either a path to a PRB file or the name of a built-in probe..
    if not path.exists():
        path = Path(__file__).parent / ('probes/%s.prb' % name_or_path)
    if not path.exists():
        raise IOError("The probe `{}` cannot be found.".format(name_or_path))
    return MEA(probe=read_python(path))


def list_probes():
    """Return the list of built-in probes."""
    path = Path(__file__).parent / 'probes'
    return [fn.stem for fn in path.glob('*.prb')]


#------------------------------------------------------------------------------
# MEA class
#------------------------------------------------------------------------------

class MEA(object):
    """A Multi-Electrode Array.

    There are two modes:

    * No probe specified: one single channel group, positions and adjacency
      list specified directly.
    * Probe specified: one can change the current channel_group.

    """

    def __init__(self,
                 channels=None,
                 positions=None,
                 adjacency=None,
                 probe=None,
                 ):
        self._probe = probe
        self._channels = channels
        self._check_positions(positions)
        self._positions = positions
        # This is a mapping {channel: list of neighbors}.
        if adjacency is None and probe is not None:
            adjacency = _probe_adjacency_list(probe)
            self.channels_per_group = _channels_per_group(probe)
        self._adjacency = adjacency
        if probe:
            # Select the first channel group.
            cg = sorted(self._probe['channel_groups'].keys())[0]
            self.change_channel_group(cg)

    def _check_positions(self, positions):
        if positions is None:
            return
        positions = _as_array(positions)
        if positions.shape[0] != self.n_channels:
            raise ValueError("'positions' "
                             "(shape {0:s})".format(str(positions.shape)) +
                             " and 'n_channels' "
                             "({0:d})".format(self.n_channels) +
                             " do not match.")

    @property
    def positions(self):
        """Channel positions in the current channel group."""
        return self._positions

    @property
    def channels(self):
        """Channel ids in the current channel group."""
        return self._channels

    @property
    def n_channels(self):
        """Number of channels in the current channel group."""
        return len(self._channels) if self._channels is not None else 0

    @property
    def adjacency(self):
        """Adjacency graph in the current channel group."""
        return self._adjacency

    def change_channel_group(self, group):
        """Change the current channel group."""
        assert self._probe is not None
        self._channels = _probe_channels(self._probe, group)
        self._positions = _probe_positions(self._probe, group)



# -*- coding: utf-8 -*-

"""Utility functions for test datasets."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import hashlib
import logging
from pathlib import Path

from phylib.utils._misc import ensure_dir_exists, phy_config_dir
from phylib.utils.event import ProgressReporter


logger = logging.getLogger(__name__)


#------------------------------------------------------------------------------
# Utility functions
#------------------------------------------------------------------------------

def _remote_file_size(path):
    import requests
    try:  # pragma: no cover
        response = requests.head(path)
        return int(response.headers.get('content-length', 0))
    except Exception:
        # Unable to get the file size: no progress report.
        pass
    return 0


def _save_stream(r, path):
    size = _remote_file_size(r.url)
    pr = ProgressReporter()
    pr.value_max = size or 1
    pr.set_progress_message('Downloading `' + str(path) + '`: {progress:.1f}%.')
    pr.set_complete_message('Download complete.')
    downloaded = 0
    with open(path, 'wb') as f:
        for i, chunk in enumerate(r.iter_content(chunk_size=1024)):
            if chunk:
                f.write(chunk)
                f.flush()
                downloaded += len(chunk)
                if i % 100 == 0:
                    pr.value = downloaded
    pr.set_complete()


def _download(url, stream=None):
    from requests import get
    r = get(url, stream=stream)
    if r.status_code != 200:  # pragma: no cover
        logger.debug("Error while downloading %s.", url)
        r.raise_for_status()
    return r


def download_text_file(url):
    """Download a text file."""
    return _download(url).text


def _md5(path, blocksize=2 ** 20):
    """Compute the checksum of a file."""
    m = hashlib.md5()
    with open(path, 'rb') as f:
        while True:
            buf = f.read(blocksize)
            if not buf:
                break
            m.update(buf)
    return m.hexdigest()


def _check_md5(path, checksum):
    return (_md5(path) == checksum) if checksum else None


def _check_md5_of_url(output_path, url):
    try:
        checksum = download_text_file(url + '.md5').split(' ')[0]
    except Exception:
        checksum = None
    finally:
        if checksum:
            return _check_md5(output_path, checksum)


def download_file(url, output_path):
    """Download a binary file from an URL.

    The checksum will be downloaded from `URL + .md5`. If this download
    succeeds, the file's MD5 will be compared to the expected checksum.

    Parameters
    ----------

    url : str
        The file's URL.
    output_path : str
        The path where the file is to be saved.

    """
    path = Path(output_path)
    if path.exists():
        checked = _check_md5_of_url(output_path, url)
        if checked is False:
            logger.debug(
                "The file `%s` already exists but is invalid: redownloading.", output_path)
        elif checked is True:
            logger.debug("The file `%s` already exists: skipping.", output_path)
            return output_path
    r = _download(url, stream=True)
    _save_stream(r, output_path)
    if _check_md5_of_url(output_path, url) is False:
        logger.debug("The checksum doesn't match: retrying the download.")
        r = _download(url, stream=True)
        _save_stream(r, output_path)
        if _check_md5_of_url(output_path, url) is False:
            raise RuntimeError("The checksum of the downloaded file "
                               "doesn't match the provided checksum.")
    return


_BASE_URL = 'https://raw.githubusercontent.com/kwikteam/phy-data/master/'


def download_test_file(name, config_dir=None, force=False):
    """Download a test file."""
    config_dir = Path(config_dir or phy_config_dir())
    path = config_dir / 'test_data' / name
    # Ensure the directory exists.
    ensure_dir_exists(path.parent)
    if not force and path.exists():
        return path
    url = _BASE_URL + name
    download_file(url, output_path=path)
    return path



# -*- coding: utf-8 -*-

"""Test MEA."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

from pathlib import Path

from pytest import raises
import numpy as np
from numpy.testing import assert_array_equal as ae

from ..mea import (_probe_channels, _remap_adjacency, _adjacency_subset,
                   _probe_positions, _probe_adjacency_list,
                   MEA, load_probe, list_probes
                   )


#------------------------------------------------------------------------------
# Tests
#------------------------------------------------------------------------------

def test_remap():
    adjacency = {1: [2, 3, 7], 3: [5, 11]}
    mapping = {1: 3, 2: 20, 3: 30, 5: 50, 7: 70, 11: 1}
    remapped = _remap_adjacency(adjacency, mapping)
    assert sorted(remapped.keys()) == [3, 30]
    assert remapped[3] == [20, 30, 70]
    assert remapped[30] == [50, 1]


def test_adjacency_subset():
    adjacency = {1: [2, 3, 7], 3: [5, 11], 5: [1, 2, 11]}
    subset = [1, 5, 32]
    adjsub = _adjacency_subset(adjacency, subset)
    assert sorted(adjsub.keys()) == [1, 5]
    assert adjsub[1] == []
    assert adjsub[5] == [1]


def test_probe():
    probe = {'channel_groups': {
             0: {'channels': [0, 3, 1],
                 'graph': [[0, 3], [1, 0]],
                 'geometry': {0: (10, 10), 1: (10, 20), 3: (20, 30)},
                 },
             1: {'channels': [7],
                 'graph': [],
                 },
             }}
    adjacency = {0: set([1, 3]),
                 1: set([0]),
                 3: set([0]),
                 }
    assert _probe_channels(probe, 0) == [0, 3, 1]
    ae(_probe_positions(probe, 0), [(10, 10), (20, 30), (10, 20)])
    assert _probe_adjacency_list(probe) == adjacency

    mea = MEA(probe=probe)

    assert mea.adjacency == adjacency
    assert mea.channels_per_group == {0: [0, 3, 1], 1: [7]}
    assert mea.channels == [0, 3, 1]
    assert mea.n_channels == 3
    ae(mea.positions, [(10, 10), (20, 30), (10, 20)])


def test_mea():

    n_channels = 10
    channels = np.arange(n_channels)
    positions = np.random.randn(n_channels, 2)

    mea = MEA(channels, positions=positions)
    ae(mea.positions, positions)
    assert mea.adjacency is None

    mea = MEA(channels, positions=positions)
    assert mea.n_channels == n_channels

    mea = MEA(channels, positions=positions)
    assert mea.n_channels == n_channels

    with raises(ValueError):
        MEA(channels=np.arange(n_channels + 1), positions=positions)

    with raises(ValueError):
        MEA(channels=channels, positions=positions[:-1, :])


def test_library(tempdir):
    assert '1x32_buzsaki' in list_probes()

    probe = load_probe('1x32_buzsaki')
    assert probe
    assert probe.channels == list(range(32))

    path = Path(tempdir) / 'test.prb'
    with raises(IOError):
        load_probe(path)

    with open(path, 'w') as f:
        f.write('')
    with raises(KeyError):
        load_probe(path)



# -*- coding: utf-8 -*-

"""Tests of misc utility functions."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import numpy as np
from numpy.testing import assert_array_equal as ae
from pytest import raises, mark

from .._misc import (
    _git_version, load_json, save_json, load_pickle, save_pickle, read_python, write_python,
    read_text, write_text, _read_tsv_simple, _write_tsv_simple, read_tsv, write_tsv,
    _pretty_floats, _encode_qbytearray, _decode_qbytearray, _fullname, _load_from_fullname)


#------------------------------------------------------------------------------
# Misc tests
#------------------------------------------------------------------------------

def test_qbytearray(tempdir):
    try:
        from PyQt5.QtCore import QByteArray
    except ImportError:  # pragma: no cover
        return
    arr = QByteArray()
    arr.append('1')
    arr.append('2')
    arr.append('3')

    encoded = _encode_qbytearray(arr)
    assert isinstance(encoded, str)
    decoded = _decode_qbytearray(encoded)
    assert arr == decoded

    # Test JSON serialization of QByteArray.
    d = {'arr': arr}
    path = tempdir / 'test'
    save_json(path, d)
    d_bis = load_json(path)
    assert d == d_bis


def test_pretty_float():
    assert _pretty_floats(0.123456) == '0.12'
    assert _pretty_floats([0.123456]) == ['0.12']
    assert _pretty_floats({'a': 0.123456}) == {'a': '0.12'}


def test_json_simple(tempdir):
    d = {'a': 1, 'b': 'bb', 3: '33', 'mock': {'mock': True}}

    path = tempdir / 'test_dir/test'
    save_json(path, d)
    d_bis = load_json(path)
    assert d == d_bis

    path.write_text('')
    assert load_json(path) == {}
    with raises(IOError):
        load_json('%s_bis' % path)


@mark.parametrize('kind', ['json', 'pickle'])
def test_json_numpy(tempdir, kind):
    arr = np.arange(20).reshape((2, -1)).astype(np.float32)
    d = {'a': arr, 'b': arr.ravel()[:10], 'c': arr[0, 0]}

    path = tempdir / 'test'
    f = save_json if kind == 'json' else save_pickle
    f(path, d)

    f = load_json if kind == 'json' else load_pickle
    d_bis = f(path)
    arr_bis = d_bis['a']

    assert arr_bis.dtype == arr.dtype
    assert arr_bis.shape == arr.shape
    ae(arr_bis, arr)

    ae(d['b'], d_bis['b'])
    ae(d['c'], d_bis['c'])


def test_read_python(tempdir):
    path = tempdir / 'mock.py'
    with open(path, 'w') as f:
        f.write("""a = {'b': 1}""")

    assert read_python(path) == {'a': {'b': 1}}


def test_write_python(tempdir):
    data = {'a': 1, 'b': 'hello', 'c': [1, 2, 3]}
    path = tempdir / 'mock.py'

    write_python(path, data)
    assert read_python(path) == data


def test_write_text(tempdir):
    for path in (tempdir / 'test_1',
                 tempdir / 'test_dir/test_2.txt',
                 ):
        write_text(path, 'hello world')
        assert read_text(path) == 'hello world'


def test_write_tsv_simple(tempdir):
    path = tempdir / 'test.tsv'
    assert _read_tsv_simple(path) == {}

    # The read/write TSV functions conserve the types: int, float, or strings.
    data = {2: 20, 3: 30.5, 5: 'hello'}
    _write_tsv_simple(path, 'myfield', data)

    assert _read_tsv_simple(path) == ('myfield', data)


def test_write_tsv(tempdir):
    path = tempdir / 'test.tsv'
    assert read_tsv(path) == []
    write_tsv(path, [])

    data = [{'a': 1, 'b': 2}, {'a': 10}, {'b': 20, 'c': 30.5}]

    write_tsv(path, data)
    assert read_tsv(path) == data

    write_tsv(path, data, first_field='b', exclude_fields=('c', 'd'))
    assert read_text(path)[0] == 'b'
    del data[2]['c']
    assert read_tsv(path) == data


def test_git_version():
    v = _git_version()
    assert v


def _myfunction(x):
    return


def test_fullname():
    assert _fullname(_myfunction) == 'phylib.utils.tests.test_misc._myfunction'

    assert _load_from_fullname(_myfunction) == _myfunction
    assert _load_from_fullname(_fullname(_myfunction)) == _myfunction



# -*- coding: utf-8 -*-

"""Tests of CCG functions."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import numpy as np
from numpy.testing import assert_array_equal as ae

from ..ccg import (_increment,
                   _diff_shifted,
                   correlograms,
                   firing_rate,
                   )


#------------------------------------------------------------------------------
# Tests
#------------------------------------------------------------------------------

def _random_data(max_cluster):
    sr = 20000
    nspikes = 10000
    spike_samples = np.cumsum(np.random.exponential(scale=.025, size=nspikes))
    spike_samples = (spike_samples * sr).astype(np.uint64)
    spike_clusters = np.random.randint(0, max_cluster, nspikes)
    return spike_samples, spike_clusters


def _ccg_params():
    return .001, .05


def test_utils():
    # First, test _increment().

    # Original array.
    arr = np.arange(10)
    # Indices of elements to increment.
    indices = [0, 2, 4, 2, 2, 2, 2, 2, 2]

    ae(_increment(arr, indices), [1, 1, 9, 3, 5, 5, 6, 7, 8, 9])

    # Then, test _shitdiff.
    # Original array.
    arr = [2, 3, 5, 7, 11, 13, 17]
    # Shifted once.
    ds1 = [1, 2, 2, 4, 2, 4]
    # Shifted twice.
    ds2 = [3, 4, 6, 6, 6]

    ae(_diff_shifted(arr, 1), ds1)
    ae(_diff_shifted(arr, 2), ds2)


def test_firing_rate_0():
    spike_clusters = [0, 1, 0, 1]
    bin_size = 1

    fr = firing_rate(spike_clusters, bin_size=bin_size, duration=20)
    ae(fr, .2 * np.ones((2, 2)))


def test_firing_rate_1():
    spike_clusters = [0, 1, 0, 1]
    bin_size = 1

    fr = firing_rate(spike_clusters, cluster_ids=[0, 1, 2], bin_size=bin_size, duration=20)
    print(fr)
    ae(fr[:2, :2], .2 * np.ones((2, 2)))
    assert np.all(fr[2, :] == fr[:, 2])
    assert np.all(fr[2, :] == 0)


def test_firing_rate_2():
    spike_clusters = np.tile(np.arange(10), 100)
    fr = firing_rate(spike_clusters, cluster_ids=np.arange(10), bin_size=.1, duration=1.)
    ae(fr, np.ones((10, 10)) * 1000)


def test_ccg_0():
    spike_samples = [0, 10, 10, 20]
    spike_clusters = [0, 1, 0, 1]
    bin_size = 1
    winsize_bins = 2 * 3 + 1

    c_expected = np.zeros((2, 2, 4))

    # WARNING: correlograms() is sensitive to the order of identical spike
    # times. This needs to be taken into account when post-processing the
    # CCGs.
    c_expected[1, 0, 0] = 1
    c_expected[0, 1, 0] = 0  # This is a peculiarity of the algorithm.

    c = correlograms(spike_samples, spike_clusters,
                     bin_size=bin_size, window_size=winsize_bins,
                     cluster_ids=[0, 1], symmetrize=False)

    ae(c, c_expected)


def test_ccg_1():
    spike_samples = np.array([2, 3, 10, 12, 20, 24, 30, 40], dtype=np.uint64)
    spike_clusters = [0, 1, 0, 0, 2, 1, 0, 2]
    bin_size = 1
    winsize_bins = 2 * 3 + 1

    c_expected = np.zeros((3, 3, 4))
    c_expected[0, 1, 1] = 1
    c_expected[0, 0, 2] = 1

    c = correlograms(spike_samples, spike_clusters,
                     bin_size=bin_size, window_size=winsize_bins,
                     symmetrize=False)

    ae(c, c_expected)


def test_ccg_2():
    max_cluster = 10
    spike_samples, spike_clusters = _random_data(max_cluster)
    bin_size, winsize_bins = _ccg_params()

    c = correlograms(
        spike_samples, spike_clusters, bin_size=bin_size, window_size=winsize_bins,
        sample_rate=20000, symmetrize=False)

    assert c.shape == (max_cluster, max_cluster, 26)


def test_ccg_symmetry_time():
    """Reverse time and check that the CCGs are just transposed."""

    spike_samples, spike_clusters = _random_data(2)
    bin_size, winsize_bins = _ccg_params()

    c0 = correlograms(spike_samples, spike_clusters,
                      bin_size=bin_size, window_size=winsize_bins,
                      sample_rate=20000, symmetrize=False)

    spike_samples_1 = np.cumsum(np.r_[np.arange(1),
                                      np.diff(spike_samples)[::-1]])
    spike_samples_1 = spike_samples_1.astype(np.uint64)
    spike_clusters_1 = spike_clusters[::-1]
    c1 = correlograms(spike_samples_1, spike_clusters_1,
                      bin_size=bin_size, window_size=winsize_bins,
                      sample_rate=20000, symmetrize=False)

    # The ACGs are identical.
    ae(c0[0, 0], c1[0, 0])
    ae(c0[1, 1], c1[1, 1])

    # The CCGs are just transposed.
    ae(c0[0, 1], c1[1, 0])
    ae(c0[1, 0], c1[0, 1])


def test_ccg_symmetry_clusters():
    """Exchange clusters and check that the CCGs are just transposed."""

    spike_samples, spike_clusters = _random_data(2)
    bin_size, winsize_bins = _ccg_params()

    c0 = correlograms(spike_samples, spike_clusters,
                      bin_size=bin_size, window_size=winsize_bins,
                      sample_rate=20000, symmetrize=False)

    spike_clusters_1 = 1 - spike_clusters
    c1 = correlograms(spike_samples, spike_clusters_1,
                      bin_size=bin_size, window_size=winsize_bins,
                      sample_rate=20000, symmetrize=False)

    # The ACGs are identical.
    ae(c0[0, 0], c1[1, 1])
    ae(c0[1, 1], c1[0, 0])

    # The CCGs are just transposed.
    ae(c0[0, 1], c1[1, 0])
    ae(c0[1, 0], c1[0, 1])


def test_symmetrize_correlograms():
    spike_samples, spike_clusters = _random_data(3)
    bin_size, winsize_bins = _ccg_params()

    sym = correlograms(spike_samples, spike_clusters,
                       bin_size=bin_size, window_size=winsize_bins,
                       sample_rate=20000)
    assert sym.shape == (3, 3, 51)

    # The ACG are reversed.
    for i in range(3):
        ae(sym[i, i, :], sym[i, i, ::-1])

    ae(sym[0, 1, :], sym[1, 0, ::-1])
